#ifndef _LOGIC_H
#define _LOGIC_H
#define _CRT_SECURE_NO_WARNINGS

#include <time.h>
#include <list>
#include <string>
#include "Graphics.h"
#include "../Common/Query.h"

/*!
@file Logic.h
Soubor deklaruje strukturu @ref RBusPrint s lidsky itelnmi daji o autobusu, dle vtov typ chybovch hlen
a tdu @ref CLogic zajiujc pevod dotazu uivatele na dotazy pro databzi a ppadn nastaven chybov hodnoty
pro vechny nepovolen vstupy.
*/

//! informace o autobusu v lidsky iteln podob
struct RBusPrint {
  //!ID autobusu
  unsigned uiId;
  //!slo linky autobusu
  unsigned uiLine;
  //!nzev vchozho msta
  std::string strCityFrom;
  //!nzev clovho msta
  std::string strCityTo;
  //!as odjezdu v UNIX timestamp
  time_t tLeaveTime;
  //!pln autobusu
  std::string strBusPlan;
};

//! kdy chybovch hlen
enum Error {NOERR, CITYFROM, CITYTO, LEAVETIME, CTFROMNFOUND, CTTONFOUND, CTNFOUND, TIMEFORMAT, BUSNFOUND,
  NBUSESFOUND, NUMSEATS, NBUSSET, NUSERSET, NFREESEATS, WRITESEATS, NRESERVE, NUNLOCK, WRITEBLINE, WRITEBPLAN,
  PLANNFOUND, WRITEBID, NDELBUS, WRCITY, TOOMANYARGS, NADDCITY, NDELCITY, NADDALIAS, WRALIAS, WRBUstrPlan, WRSEATSCOUNT,
  NADDPLAN, NDELPLAN, NFILE, NADDBUS, WRONGBID, FIRSTFIND, WRUSERID, NTYPESET, WRARGUMENT, FILEREADFAIL, BUSDISABLED,
  USEDBUS, NDISBUS};

//! tda na pevod uivatelskch dotaz na dotazy pro databzi
/*!
Tda zajiuje pevod dotazu uivatele na dotazy pro databzi a ppadn nastaven chybov hodnoty pro vechny nepovolen vstupy.
Chybov kdy po kadm pkazu si meme zjistit volnm @ref GetLastError().
Pamatuje si pedtm uivatelem zadan hodnoty a tm umouje pouvat jednodu a snze zapamatovateln pkazy.
*/
class CLogic {
public:
  //! konstruktor
  /*!
  @param[in] cQuery ukazatel na njakou tdu poskytujc rozhran @ref CQuery, pomoc kter bude probhat komunikace s databz
  */
  CLogic(CQuery *cQuery);
  //! destruktor
  ~CLogic();
  //! inicializace, mj. pidlen uniktnho ID novmu zkaznkovi
  void Init(void);
  //! vymazn aktulnch dat
  /*!
  Vymae aktuln hodnoty jako ID aktulnho uivatele, aktuln vybran autobus, vchoz a clov msto
  z poslednho hledn a as odjezdu poslednho nalezenho autobusu.
  */
  void Clear(void);
  //! nalezen vyhovujcch autobus
  /*!
  Funkce pro nalezen danho potu autobus. Poet je dn pi pekladu v privtn poloce @e uiPrintCount.
  @param[out] cBuses list nalezench autobus k vypsn
  @param[in] pszCityFrom etzec s nzvem vchozho msta
  @param[in] pszCityTo etzec s nzvem clovho msta
  @param[in] pszLeaveTime etzec s asem, od kterho dl se m v databti vyhledvat
  @return @e true pi spchu, @e false jinak
  */
  bool Find(std::list<RBusPrint> *cBuses, const char *pszCityFrom, const char *pszCityTo, const char *pszLeaveTime);
  //! vypsn dalch autobus
  /*!
  Vypsn dalch autobus. Nutn podmnka je, aby byl ped tmto pouit pkaz "find".
  @param[out] cBuses list nalezench autobus k vypsn
  @return @e true pi spchu, @e false jinak
  */
  bool Next(std::list<RBusPrint> *cBuses);
  //! nastav aktuln autobus
  /*!
  Funkce pro nastaven autobusu se zadanm ID jako aktivnho.
  @param[in] uiBusId ID autobusu, kterho chceme nastavit jako aktivnho
  @param[out] cFreeSeats seznam volnch sedadel nastavenho autobusu
  @param[out] strPlan etzec s plnkem autobusu v textov form
  @param[out] prBus informace o lince
  @param[out] puiNumberSeats celkov poet sedadel v autobusu
  @return @e true pi spchu, @e false jinak
  */
  bool SetBus(unsigned uiBusId, std::list<unsigned> *cFreeSeats, std::string *strPlan, RBusPrint *prBus, unsigned *puiNumberSeats);
  //! doasn rezervace
  /*!
  Zarezervuje zadan poet sedadel jen na uritou dobu, pot budou automaticky uvolnna.
  @param[out] cSeats seznam sedadel, kter se podailo rezervovat
  @param[in] uiSeatsCount poet sedadel k rezervaci
  @return @e true pokud se podailo rezervovat vechna sedadla, @e false jinak
  */
  bool ReserveAsk(std::list<unsigned> *cSeats, unsigned uiSeatsCount);
  //! doasn rezervace
  /*!
  Zarezervuje zadan sedadla jen na uritou dobu, pot budou automaticky uvolnna
  @param[in,out] cSeats seznam sedadel k rezervaci. Sedadla, kter se nepovedlo rezervovat,
  budou z tohoto seznamu odstranna.
  @return @e true pokud se podailo rezervovat vechna sedadla, @e false jinak
  */
  bool ReserveAsk(std::list<unsigned> *cSeats);
  //! konen rezervace
  /*!
  Definitivn zarezervuje zadan sedadla.
  @param[in,out] cSeats seznam sedadel k rezervaci. Sedadla, kter se nepovedlo rezervovat,
    budou z tohoto seznamu odstranna.
  @param[out] puiId ID uivatele, kter provedl rezervaci
  @param[out] prBus data o autobusu, ve kterm byla provedene rezervace
  @return @e true pokud se podailo rezervovat vechna sedadla, @e false jinak
  */
  bool ReserveForever(std::list<unsigned> *cSeats, unsigned *puiId, RBusPrint *prBus);
  //! uvolnn sedadel
  /*!
  Provede uvolnn sedadel v aktulnm autobusu se zadanm ID uivatele
  @param[in] uiUserId ID uivatele, kter chce zruit svou rezervaci
  @param[out] cSeats seznam uvolnnch sedadel
  @return @e true pokud se podailo uvolnit sedadla, @e false jinak
  */
  bool UnlockAll(unsigned uiUserId, std::list<unsigned> *cSeats);
  //! uvolnn sedadel
  /*!
  Provede uvolnn zadanch sedadel v aktulnm autobusu s aktulnm ID uivatele. Uiten zejmna pro
  uvolnn doasn rezervovanch sedadel pi nepotvrzen rezervace.
  @param[in,out] cSeats seznam sedadel k uvolnn, na vstupu uvolnn sedadla
  @return @e true pokud se podailo uvolnit sedadla, @e false jinak
  */
  bool Unlock(std::list<unsigned> *cSeats);
  //! uvolnn sedadel
  /*!
  Provede uvolnn zadanch sedadel v aktulnm autobusu se zadanm ID uivatele.
  @param[in] uiUserId ID uivatele, kter chce zruit svou rezervaci
  @param[in,out] cSeats seznam sedadel k uvolnn, na vstupu uvolnn sedadla
  @return @e true pokud se podailo uvolnit sedadla, @e false jinak
  */
  bool Unlock(unsigned uiUserId, std::list<unsigned> *cSeats);
  //! pidn novho msta
  /*!
  Pid nov msto do databze.
  @param[in] pszCityName nzev nov pidvanho msta
  @return ID nov pidanho msta, v ppad nespchu 0
  */
  unsigned AddCity(const char *pszCityName);
  //! smazn msta
  /*!
  Smae stvajc msto z databze.
  @param[in] pszCityName nzev msta ke smazn
  @return @e true pi spchu, @e false jinak
  */
  bool DelCity(const char *pszCityName);
  //! pdn zkratky pro stvajc msto
  /*!
  Pidn jmna msta, kter se odkazuje na njak stvajc msto.
  @param[in] pszCityAlias nov zkratka pro msto
  @param[in] pszCityName jmno stvajcho msta
  @return @e true pi spchu, @e false jinak
  */
  bool AddCityAlias(const char *pszCityAlias, const char *pszCityName);
  //! pidn novho autobusu
  /*!
  Pid nov autobus do databze.
  @param[in] uiLine slo linky autobusu
  @param[in] pszCityFrom nzev vchozho msta
  @param[in] pszCityTo nzev clovho msta
  @param[in] pszLeaveTime etzec s asem odjezdu autobusu
  @param[in] pszPlan nzev plnu/typu autobusu
  @return ID nov pidanho autobusu, 0 v ppad chyby
  */
  unsigned AddBus(unsigned uiLine, const char *pszCityFrom, const char *pszCityTo, const char *pszLeaveTime, const char *pszPlan);
  //! smazn autobusu
  /*!
  Smae autobus podle jeho ID.
  @param[in] uiBusId ID autobusu urenho ke smazn
  @return @e true pi spchu, @e false jinak
  */
  bool DelBus(unsigned uiBusId);
  //! pidn novho plnu autobusu
  /*!
  Pid do databze nov pln i typ autobusu.
  @param[in] pszName nzev plnu autobusu
  @param[in] uiSeats poet sedadel v tomto plnu
  @param[in] pszFile nzev textovho souboru obsahujcho pln v textov form
  @return ID pidanho plnu, 0 v ppad chyby
  */
  unsigned AddBusPlan(const char *pszName, unsigned uiSeats, const char *pszFile);
  //! smazn plnu autobusu
  /*!
  Smae pln autobusu z databze.
  @param[in] pszName jmno plnu autobusu ke smazn
  @return @e true pi spchu, @e false jinak
  */
  bool DelBusPlan(const char *pszName);
  //! uzamknut autobusu
  /*
  Uzamkne autobus pro dal rezervace a vype seznam obsazench sedadel.
  @param[in] uiBusId slo autobusu
  @param[out] cReservedSeats list obsazench sedadel
  @param[out] strPlan plnek autobusu
  @param[out] prBusPrint informace o autobusu
  @param[out] puiSeats poet sedadel autobusu
  @return @e true pi spchu, @e false jinak
  */
  bool Close(unsigned uiBusId, std::list<unsigned> *cReservedSeats, std::string *strPlan, RBusPrint *prBusPrint, unsigned *puiSeats);
  //! zjitn ppadn chyby v posledn funkci
  /*!
  Zjist stav, se kterm skonila posledn volan funkce.
  @return chybov stav z @ref Error
  */
  Error GetLastError(void) const;
  //! smazn starch autobus
  /*!
  Smae autobusy, kter ji odjely.
  @return poet smazanch autobus
  */
  unsigned ClearBus(void);
  //! smazn nepouvanch mst
  /*!
  Smae msta, do kterch ani z kterch nejezd dn autobusy.
  @return poet smazanch mst
  */
  unsigned ClearCity(void);
  //! smazn nepouvanch pln autobus
  /*!
  Smae nevyuvan plny autobus z databze.
  @return poet smazanch pln
  */
  unsigned ClearPlan(void);
  //! vypsn existujcch mst
  /*!
  Vype existujc msta z databze.
  @param[in] in vstupn sekvence, podle kter se hledaj podobn msta
  @param[out] cOut seznam nalezench mst. Nzvy mst maj druh string przdn, aliasy jsou v druhm
  stringu a v prvnm je i opravdov nzev msta.
  */
  void City(std::string& in, std::vector<std::pair<std::string, std::string>>& cOut);
private:
  static const unsigned uiPrintCount = 3;
  CQuery *m_cDatabase;
  unsigned m_uiCurBus;
  unsigned m_uiCurUserId;
  unsigned m_uiCurCityFrom;
  unsigned m_uiCurCityTo;
  time_t m_tCurLeaveTime;
  time_t m_tNextLeaveTime;
  bool ReservedSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned *puiSeats = NULL);
  time_t ConvertTime(const char *pszTime);
  Error m_eLastError;
};


#endif
