#ifndef _QUERY_H
#define _QUERY_H

#include <string>
#include <list>
#include <time.h>
#include "UtilClass.h"

/*!
@file Query.h
Poskytuje tdu @ref CQuery s ist virtulnmi metodami, kter zasteuje rozhran mezi databz a zbytkem programu.
Kad ppadn dal implementace databze mus dodret stejn rozhran, kter by mla zddit od @ref CQuery.
Dle zde najdeme jednoduch struktury uren pro snadn pedvn dat mezi funkcemi.
*/

//! Informace o mst
struct RCityInfo {
  //! ID msta
  unsigned uiId;
  //! nzev msta
  std::string strName;
};

//! Informace o plnu autobusu
struct RPlanInfo {
  //! ID plnu
  unsigned uiId;
  //! nzev plnu nebo typu autobusu
  std::string strName;
  //! poet sedadel
  unsigned uiSeats;
};

 //! Informace o autobusu
struct RBusInfo {
  //! ID autobusu
  unsigned uiId;
  //! slo linky autobusu
  unsigned uiLine;
  //! slo potenho msta
  unsigned uiCityFrom;
  //! slo clovho msta
  unsigned uiCityTo;
  //! slo plnu
  unsigned uiPlan;
  //! as odjezdu z vchoz stanice
  time_t tLeaveTime;
};

//! Tda urujc klientsk rozhran mezi databz a zbytkem programu
/*!
Tato tda poskytuje pouze seznam ist virtulnch metod, kter je teba implementovat pro kadou databzi,
se kterou budeme chtt pracovat. Kad implementace databze by mla bt potomkem tto tdy, nicmn rozhran 
samo nijak neuruje detaily ani pro databzov stroj ani pro uloen dat.
*/
class CQuery {
public:
  //! destruktor
  /*!
  Destruktor je deklarovn jako virtuln, sm neprovd dnou innost.
  */
  virtual ~CQuery() = 0;
  //! pidn msta
  /*!
  @param[in] pszName jmno msta
  @return uniktn slo nov pidanho msta i slo ji dve existujcho msta
  */
  virtual unsigned AddCity(const char *pszName) = 0;
  //! pidn zkratky k njakmu existujcmu mstu
  /*!
  @param[in] pszName nzev zkratky i dalho nzvu
  @param[in] uiId ID existujcho msta, se kterm chceme propojit tuto zkratku
  @return spch operace - @e false v ppad, e jmno ji existuje (jako msto i alias) nebo
  pokud msto s ID nen v databzi
  */
  virtual bool AddCityAlias(const char *pszName, unsigned uiId) = 0;
  //! smazn msta podle jmna
  /*!
  Provede smazn msta z kontajneru a zrove vymazn vech ppadnch alias nzvu msta.
  @param[in] pszName nzev i alias msta
  @return spch operace - @e false v ppad, e msto nen v databzi nebo existuj autobusov
  linky, kter msto pouvaj
  */
  virtual bool DeleteCity(const char *pszName) = 0;
  //! nalezen msta podle nzvu
  /*!
  Hledn msta probh nejprve v normlnch nzvech mst a v ppad nespchu ve zkratkch (aliasech)
  @param[in] pszName nzev i alias msta
  @return ID nalezenho msta nebo 0 v ppad nespchu
  */
  virtual unsigned FindCity(const char *pszName) = 0;
  //! zskn informac o mst
  /*!
  @param[in] uiCity ID msta
  @param[out] prInfo struktura @ref RCityInfo k vyplnn daty
  @param[out] puiLinesOut poet autobus odjdjcch z tohoto msta
  @param[out] puiLinesIn poet autobus pijdjcch do tohoto msta
  @return spch operace - @e false, pokud msto nebylo nalezeno
  @note Vstupn parametry mohou bt i @e NULL.
  */
  virtual bool CityInfo(unsigned uiCity, RCityInfo *prInfo, unsigned *puiLinesOut, unsigned *puiLinesIn) = 0;
  //! pidn plnu autobusu
  /*!
  @param[in] pszName nzev plnu/typu autobusu
  @param[in] uiSeats poet sedadel
  @param[in] cData binrn data rozvren sedadel
  @return uniktn slo pidanho plnu nebo slo ji existujcho plny se zadanm jmnem 
  */
  virtual unsigned AddBusPlan(const char *pszName, unsigned uiSeats, const CBlobEx<std::string> *cData) = 0;
  //! smazn plnu autobusu
  /*!
  Vymazn plnu autobusu z kontajneru, me bt provedeno pouze v ppad, e neexistuje dn autobus
  s tmto plnem.
  @param[in] pszName nzev plnu autobusu
  @return spch operace
  */
  virtual bool DeleteBusPlan(const char *pszName) = 0;
  //! nalezen plnu autobusu
  /*!
  Hledn plnu autobusu podle zadanho jmna.
  @param[in] pszName nzev plnu autobusu
  @return ID plnu, 0 v ppad nespchu
  */
  virtual unsigned FindBusPlan(const char *pszName) = 0;
  //! zskn informac o plnu
  /*!
  @param[in] uiPlan ID plnu autobusu
  @param[out] prInfo struktura @ref RPlanInfo naplnn daty
  @param[out] puiBuses poet autobus vyuvajcch tento pln
  @param[out] cData binrn data s rozloenm sedadel
  @return spch operace - @e false, pokud pln nebyl nalezen
  @note Vstupn parametry mohou bt i @e NULL.
  */
  virtual bool BusPlanInfo(unsigned uiPlan, RPlanInfo *prInfo, unsigned *puiBuses, CBlobEx<std::string> *cData) = 0;
  //! pidn autobusu
  /*!
  Vytvoen autobusu s uvedenmi daty a jeho pidn do databze. Identifikan slo 
  (poloka @c uiId struktury @c RBusInfo) nemus bt vyplnno, nebo dojde k jeho naplnn
  jedinenm slem a pi zpracovn poadavku.
  @param[in] prBus struktura @ref RBusInfo s daty potebnmi k vytvoen novho autobusu,
  @return ID nov vloenho autobusu nebo 0 v ppad chyby (snaha o zaveden duplicitnho
  autobusu)
  */
  virtual unsigned AddBus(const RBusInfo *prBus) = 0;
  //! smazn autobusu
  /*!
  @param[in] uiBus ID autobusu urenho ke smazn
  @return spch operace
  */
  virtual bool DeleteBus(unsigned uiBus) = 0;
  //! nalezen autobusu
  /*!
  Funkce pro nalezen autobusu z msta @a uiFrom do msta @a uiTo a s asem odjezdu rovnm nebo
  nejblim vtm asu @a tLeaveTime.
  @param[in] uiFrom ID vchozho msta
  @param[in] uiTo ID clovho msta
  @param[in] tLeaveTime as, od kterho hledme shodu (tj. doln hranice asu odjezdu)
  @return ID nalezenho autobusu nebo 0
  */
  virtual unsigned FindBus(unsigned uiFrom, unsigned uiTo, time_t tLeaveTime) = 0;
  //! zskn informac o autobusu
  /*!
  @param[in] uiBus ID autobusu
  @param[out] prInfo struktura @ref RBusInfo, kter bude naplnn daty
  @param[out] pfEnabled pepna, zda jsou povoleny zmny v autobusu
  @param[out] puiSeats poet sedadel autobusu
  @param[out] cFreeSeats seznam volnch sedadel
  @return spch operace - @e false, nebyl-li autobus nalezen
  @note Vstupn parametry mohou bt i @e NULL.
  */
  virtual bool BusInfo(unsigned uiBus, RBusInfo *prInfo, bool *pfEnabled, unsigned *puiSeats, std::list<unsigned> *cFreeSeats) = 0;
  //! povolen autobusu
  /*!
  Povol provdt zmny v rezervacch.
  @param[in] uiBus ID autobusu
  @return spch operace - @e false, nebyl-li autobus nalezen
  */
  virtual bool EnableBus(unsigned uiBus) = 0;
  //! zakzn autobusu
  /*!
  Zake provdt zmny v rezervacch.
  @param[in] uiBus ID autobusu
  @return spch operace - @e false, nebyl-li autobus nalezen
  */
  virtual bool DisableBus(unsigned uiBus) = 0;
  //! uniktn ID pro novho zkaznka
  /*!
  Operace rezervace sedadel je vdy vzna na konkrtnho zkaznka identifikovanho jedinenm slem.
  Dky tomu pak me zkaznk sv rezervace dodaten zruit nebo potvrdit doasnou rezervaci.
  @return uniktn slo zkaznka
  */
  virtual unsigned NewCustomer(void) = 0;
  //! doasn zamluven sedadel
  /*!
  Doasn zamluven sedadel, kter po uritm ase vypr.
  @param[in] uiBus ID autobusu
  @param[in,out] cSeats seznam sedadel k rezervaci, nespn sedadla budou smazna ze seznamu
  @param[in] uiCustomer ID zkaznka
  @return spch operace - @e false, pokud autobus nen nalezen nebo se operace nezdaila pro
  vechna sedadla
  */
  virtual bool AskSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer) = 0;
  //! trval rezervace sedadel
  /*!
  @param[in] uiBus ID autobusu
  @param[in,out] cSeats seznam sedadel k rezervaci, nespn sedadla budou smazna ze seznamu
  @param[in] uiCustomer ID zkaznka
  @return spch operace - @e false, pokud autobus nen nalezen nebo se operace nezdaila pro
  vechna sedadla
  */
  virtual bool LockSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer) = 0;
  //! zruen rezervace sedadel
  /*!
  @param[in] uiBus ID autobusu
  @param[in,out] cSeats seznam sedadel k uvolnn, nespn sedadla budou smazna ze seznamu
  @param[in] uiCustomer ID zkaznka
  @return spch operace - @e false, pokud autobus nen nalezen nebo se operace nezdaila pro
  vechna sedadla
  @warning K spnmu zruen rezervace je nutn mt zkaznick ID, se kterm byla provedena
  rezervace.
  */
  virtual bool UnlockSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer) = 0;
  //! nalezen dalho autobusu v poad
  /*!
  Najde dal autobus podle ID.
  @param[in] uiBus ID autobusu, od kterho hledme nsledujc
  @return ID nalezenho autobusu nebo @e 0

  Pomoc tto metody je mon sekvenn prochzet kontajner autobus (zapouzden ve td 
  @ref CBuses). Odpovdajc kousek kdu me vypadat nap. nsledovn
  @code
    unsigned uiId = 0;
    while((uiId = cQuery->NextBus(uiId)) != 0) {
      // zpracovn autobusu s slem uiId
    }
  @endcode
  Metodu meme vyut nap. v situaci, kdy potebujeme zakzat rezervace ve vech autobusech,
  kter odjdj bhem nsledujc hodiny (abychom mohli pedat idim konen seznam 
  rezervovanch sedadel).
  */
  virtual unsigned NextBus(unsigned uiBus) = 0;
  //! nalezen dalho msta v poad
  /*!
  Najde dal msto podle ID.
  @param[in] uiCity ID msta, od kterho hledme nsledujc
  @return ID nalezenho msta nebo @e 0
  */
  virtual unsigned NextCity(unsigned uiCity) = 0;
  //! nalezen dalho plnu v poad
  /*!
  Najde dal pln podle ID.
  @param[in] uiPlan ID plnu, od kterho hledme nsledujc
  @return ID nalezenho plnu nebo @e 0
  */
  virtual unsigned NextPlan(unsigned uiPlan) = 0;
  //! nalezen dalho msta v abecednm poad
  /*!
  Najde dal msto v abecednm poad.
  @param[in] pszFrom nzev msta, od kterho hledme nsledujc
  @param[out] strName jmno dalho msta v poad
  @return spch operace (@e false pro dosaen konce)
  */
  virtual bool ListCity(const char *pszFrom, std::string *strName) = 0;
  //! nalezen dalho aliasu msta v abecednm poad
  /*!
  Najde dal alias v abecednm poad.
  @param[in] pszFrom alias, od kterho hledme nsledujc
  @param[out] strAlias alias dalho msta v poad
  @param[out] strCity jmno msta odpovdajc aliasu
  @return spch operace (@e false pro dosaen konce)
  */
  virtual bool ListAlias(const char *pszFrom, std::string *strAlias, std::string *strCity) = 0;
};

inline CQuery::~CQuery() {}

#endif
