#ifndef _UTILCLASS_H
#define _UTILCLASS_H

#include <windows.h>

/*!
@file UtilClass.h
Pomocn tdy pro veobecn pouit.
*/

//=========================================================================================
//! tda pro schovu pole binrnch dat
/*!
Tda slou k jednoduch manipulaci s polem binrnch dat ppadn i vt dlky. Instance
mohou slouit i jako prvky kontajner STL.
*/
class CBlob {
public:
  //! defaultn konstruktor
  /*! Datov pole je vytvoeno jako przdn. */
  CBlob(void);
  //! konstruktor s inicializac dat
  /*! 
  Hodnoty datovho pole jsou nastaveny na zklad parametr.
  @param[in] pData ukazatel na pole zdrojovch dat (me bt i @e NULL)
  @param[in] uiLength velikost pole zdrojovch dat
  */
  CBlob(const void *pData, unsigned uiLength);
  //! konstruktor s inicializac dat
  /*!
  Hodnoty datovho pole jsou nastaveny na zklad ukazatele na jinou instanci.
  @param[in] cSrc ukazatel na zdrojovou instanci (me bt i @e NULL)
  */
  CBlob(const CBlob *cSrc);
  //! koprovac konstruktor
  /*! Provd hlubokou kopii zdrojovch dat, poteba pro pouit tdy jako poloky kontajner STL. */
  CBlob(const CBlob &cSrc);
  //! destruktor
  virtual ~CBlob();
  //! naplnn dat
  /*!
  Nastaven dat, pvodn hodnoty jsou ztraceny.
  @param[in] pData ukazatel na pole zdrojovch dat (me bt i @e NULL)
  @param[in] uiLength velikost pole zdrojovch dat
  */
  void Set(const void *pData, unsigned uiLength);
  //! vyprzdnn dat
  /*! Provede uvolnn prostoru pro data a jejich reinicializaci na nulovou dlku. */
  void Clear(void);
  //! velikost uloench dat
  /*! 
  Zskn velikosti uloench dat, nap. pro alokaci dostatenho prostoru 
  pro jejich uloen.
  @return velikost v bytech
  */
  unsigned DataLen(void) const;
  //! koprovn dat
  /*!
  Proveden kopie vnitnch dat tdy do pedanho prostoru.
  @param[out] pBuffer clov adresa pro uloen dat
  @param[in] uiLength velikost clovho pole, tj. maximln velikost koprovanch dat
  @param[in] uiOffset ofset (v bytech) pro zatek koprovn
  @return velikost zkoprovanch dat v bytech
  */
  unsigned CopyData(void *pBuffer, unsigned uiLength, unsigned uiOffset = 0) const;
  //! odkaz na data
  /*! 
  Zskn ukazatele na data uloen v rmci tdy, nap. jako zdrojov adresy pro koprovn.
  @return ukazatel na data
  */
  const void *DataPointer(void) const;
  //! opertor piazen
  /*! Piazen provede sprvn i hlubokou kopii dat. */
  CBlob &operator=(const CBlob &cRight);
private:
  unsigned char *m_pbData;
  unsigned m_uiLength;
};

//=========================================================================================
//! ablona tdy pro pole binrnch dat a dal prvek
/*!
Pomoc tto ablony lze snadno vytvet instance td, kter zapouduj pole binrnch dat
(zddno od @ref CBlob) a njak dal prvek (i prvky nap. jako poloky struktury).

Pro typov parametr ablony @c T se pedpokld, e podporuje opertor piazen. Vyuvaj se
i jeho defaultn a koprovac konstruktory. Dky implementaci opertoru piazen v @ref CBlob
pak sprvn probh i piazen odvozench td.
*/
template <typename T> class CBlobEx : public CBlob {
public:
  //! defaultn konstruktor
  /*! Inicializuje poloky "nulovmi" hodnotami. */
  CBlobEx(void);
  //! konstruktor s inicializac
  /*! 
  Vnitn data jsou inicializovna pedanmi hodnotami.
  @param[in] tData hodnota bude piazena do intern poloky typu @c T
  @param[in] pData pole binrnch dat (me bt i @e NULL)
  @param[in] uiLength velikost binrnch dat
  */
  CBlobEx(T tData, const void *pData, unsigned uiLength);
  //! konstruktor s inicializac dat
  /*!
  Hodnoty datovho pole jsou nastaveny na zklad ukazatele na jinou instanci.
  @param[in] cSrc ukazatel na zdrojovou instanci (me bt i @e NULL)
  */
  CBlobEx(const CBlobEx<T> *cSrc);
  //! koprovac konstruktor
  /*! Provd kopii zdrojovch dat, u binrnch dat hlubokou. */
  CBlobEx(const CBlobEx<T> &cSrc);
  //! piazen hodnoty prvku
  /*!
  @param[in] tData hodnota bude piazena do intern poloky typu @c T
  */
  void Set(T tData);
  //! piazen hodnot
  /*!
  Vnitn data jsou aktualizovna podle pedanch hodnot.
  @param[in] tData hodnota bude piazena do intern poloky typu @c T
  @param[in] pData pole binrnch dat (me bt i @e NULL)
  @param[in] uiLength velikost binrnch dat
  */
  void Set(T tData, const void *pData, unsigned uiLength);
  //! zskn hodnoty prvku
  /*! @return hodnota prvku typu @c T*/
  T Get(void) const;
private:
  T m_tData;
};

template <typename T> 
CBlobEx<T>::CBlobEx(void) : CBlob(), m_tData() {}

template <typename T> 
CBlobEx<T>::CBlobEx(T tData, const void *pData, unsigned uiLength) : CBlob(pData, uiLength), m_tData(tData) {}

template <typename T> 
CBlobEx<T>::CBlobEx(const CBlobEx<T> *cSrc) : CBlob(), m_tData() {if(cSrc != NULL) Set(cSrc->Get(), cSrc->DataPointer(), cSrc->DataLen());}

template <typename T> 
CBlobEx<T>::CBlobEx(const CBlobEx<T> &cSrc) : CBlob(static_cast<const CBlob &>(cSrc)), m_tData(cSrc.m_tData) {}

template <typename T>
void CBlobEx<T>::Set(T tData) {m_tData = tData;}

template <typename T>
void CBlobEx<T>::Set(T tData, const void *pData, unsigned uiLength) {m_tData = tData; CBlob::Set(pData, uiLength);}

template <typename T>
T CBlobEx<T>::Get(void) const {return m_tData;}

//=========================================================================================
//! synchronizace pstupu typu vce ten - jeden zpis
/*!
Tda implementuje zmek pro synchronizaci pstupu k njakmu objektu souasn z vce vlken.
Je mon provdt zrove vce operac ten (tj. operac bez modifikace dat chrnnho objektu),
ale pouze jedinou operaci zpisu (modifikaci dat objektu). Jinmi slovy umouje sdlen (pouze
pro ten) nebo uniktn pstup (pro modifikaci) k datm. Maximln poet souasn pistupujcch 
ten se nastavuje v konstruktoru.

Zabezpeen pomoc tto tdy m smysl u pstupu k objektm, kdy operace ten je mnohem etnj
ne operace zpisu. Implementace zrove upednostuje zpis v tom smyslu, e pi poadavku na
pstup v zapisovacm reimu nejsou povolovny dal poadavky na ten a do doby ukonen 
zapisovac operace.

Pedpokld se, e potomci tto tdy mohou bt umstni v kontajnerech STL, proto je implementovn
i koprovac konstruktor a opertor piazen. Kad objekt (i nov vytvoen) si ale vytv
individuln zabezpeen, kter nen mon koprovat.

Pvodn byla navrena implementace vyuvajc objekty jdra - udlost a semafor, veejn st 
deklarace tdy a tm i pouit jsou shodn. 
@code
class CReadWriteLock {
public:
  CReadWriteLock(unsigned uiMaxReaders);
  CReadWriteLock(const CReadWriteLock &cSource);
  ~CReadWriteLock();
  void EnterRead(void);
  void LeaveRead(void);
  void EnterWrite(void);
  void LeaveWrite(void);
  CReadWriteLock &operator=(const CReadWriteLock &cRight);
private:
  unsigned m_uiMaxReaders;
  HANDLE m_ahLocks[2]; // 0 - udlost pro zpis (vt priorita), 1 - semafor pro ten
};

CReadWriteLock::CReadWriteLock(unsigned uiMaxReaders)
{
  m_uiMaxReaders = uiMaxReaders;
  // pstup pro zpis - signalizovan auto-reset udlost
  m_ahLocks[0] = CreateEvent(NULL, FALSE, TRUE, NULL); // neprobh zpis
  // pstup pro ten - semafor nastaven na max. hodnotu
  m_ahLocks[1] = CreateSemaphore(NULL, uiMaxReaders, uiMaxReaders, NULL); // dn aktivn ten
}

CReadWriteLock::CReadWriteLock(const CReadWriteLock &cSource)
{
  m_uiMaxReaders = cSource.m_uiMaxReaders;
  // objekty jdra si vytv sv
  m_ahLocks[0] = CreateEvent(NULL, FALSE, TRUE, NULL);
  m_ahLocks[1] = CreateSemaphore(NULL, m_uiMaxReaders, m_uiMaxReaders, NULL);
}

CReadWriteLock::~CReadWriteLock()
{
  EnterWrite();
  // dn ten ani zpis nejsou aktivn - lze ruit objekty jdra
  CloseHandle(m_ahLocks[0]);
  CloseHandle(m_ahLocks[1]);
}

void CReadWriteLock::EnterRead(void)
{
  // ekn na ukonen zpisu i volnou pozici pro ten zrove
  WaitForMultipleObjects(2, m_ahLocks, TRUE, INFINITE);
  SetEvent(m_ahLocks[0]); // uvolnn omezen pro zpis
}

void CReadWriteLock::LeaveRead(void)
{
  // ten dokoneno - uvolnn pozice
  ReleaseSemaphore(m_ahLocks[1], 1, NULL);
}

void CReadWriteLock::EnterWrite(void)
{
  unsigned uiLastCount;

  // ekn na ukonen jinho zpisu
  WaitForSingleObject(m_ahLocks[0], INFINITE); // zablokuje dal EnterRead
  // ekn na dokonen vech ten
  while(true) {
    WaitForSingleObject(m_ahLocks[1], INFINITE); // -1
    ReleaseSemaphore(m_ahLocks[1], 1, (LPLONG)&uiLastCount); // +1
    if(uiLastCount == m_uiMaxReaders - 1)
      return; // ta na maximln hodnot => dn aktivn ten
    Sleep(0);
  }
}

void CReadWriteLock::LeaveWrite(void)
{
  // zpis dokonen - povolen dalho zpisu
  SetEvent(m_ahLocks[0]);
}

CReadWriteLock &CReadWriteLock::operator=(const CReadWriteLock &cRight)
{
  // dn innost
  return *this;
}
@endcode
Vzhledem k monmu vyerpn dostupnch objekt jdra (pi rozshlejch databzch) i 
s pihldnutm k vkonu je nyn tda implementovna zcela v uivatelskm prostoru. Dal
monost by bylo pouit <i>slim read-write locks</i>, kde vak nen zaruena pednost
zapisovacch poadavk (dnes ji asi nen pli podstatn, e nejsou podporovny ve
Windows XP).

@warning Pi pouit tdy je zsadn sprvn provat metody pro vstup do chrnnho bloku a 
jeho oputn! Je mon vnoen blok, ale nen sprvn takto chrnn bloky pekrvat. 
Nerespektovn tchto pravidel povede nejspe k "deadlocku" i zhroucen aplikace.
*/
class CReadWriteLock {
public:
  //! konstruktor
  /*!
  Vytvoen instance tdy a inicializace pouvanch objekt.
  @param[in] uiMaxReaders nastaven potu ten, kte mohou zrove pistupovat k chrnnmu objektu
  */
  CReadWriteLock(unsigned uiMaxReaders);
  //! koprovac konstruktor
  /*! Vytvoen instance tdy s tm, e objekty k ochran si vytv vlastn. */
  CReadWriteLock(const CReadWriteLock &cSource);
  //! destruktor
  /*! Dojde k uvolnn pouitch objekt. */
  ~CReadWriteLock();
  //! vstup do chrnnho bloku pro ten
  /*!
  Chrnn blok je zpstupnn za podmnek, e nen poadovn i neprobh zpis a zrove nen
  vyerpn povolen poet ten. Do splnn podmnek je volajc vlkno blokovno.
  */
  void EnterRead(void);
  //! oputn chrnnho bloku po ten
  /*!
  Provede uvolnn pozice pro dalho tene.
  */
  void LeaveRead(void);
  //! vstup do chrnnho bloku pro zpis
  /*!
  Chrnn blok je zpstupnn za podmnek, e nen poadovn i neprobh jin zpis a zrove nen
  aktivn dn ten.
  */
  void EnterWrite(void);
  //! oputn chrnnho bloku po zpisu
  /*!
  Chrnn blok je zpstupnn dalm operacm, a ji se jedn o ten nebo o zpis.
  */
  void LeaveWrite(void);
  //! piazovac opertor
  /*!
  Operace piazen neprovd dnou innost krom definovn nvratov hodnoty. Clov tda m
  ji z konstruktoru inicializovny vlastn prostedky pro svou innost a nen mon tyto njak navzjem
  msit. Piazovn instanc me mt smysl v ppad potomk, piem defaultn piazovac opertor
  by naruoval innost blokovacho mechanismu.
  */
  CReadWriteLock &operator=(const CReadWriteLock &cRight);
private:
  // maximln poet ten
  unsigned m_uiMaxReaders;
  // poet aktulnch ten
  unsigned m_uiReaders;
  // poet poadavk na zapisovac pstup (vetn ppadnho aktivnho zapisovatele)
  unsigned m_uiWriters;
  // pznak aktivnho zapisovatele
  bool m_fWriter;
  // ochrana konzistence privtnch promnnch
  CRITICAL_SECTION m_rLock;
};

//=========================================================================================
//! potadlo
/*!
Tda zajiuje produkovn vzrstajc ady sel typu @c unsigned. Je zajitna bezpenost
pi vcevlknovm pouit. Rozsah je dn pouitm datovm typem. Zd se, e pro ely 
jedinenho slovn zprv pi penosu, transakc, klient apod. je souasn rozsah vce ne 
dostaten. V ppad poteby je mon velikost typu pmoae zvtit.

Tda deklaruje (ale neimplementuje!) privtn koprovac konstruktor a opertor piazen. 
Tm brn veejnmu vytven dalch instanc, m by byla naruena podmnka jedinenho slovn.
*/
class CAutoIncrement {
public:
  //! konstruktor
  /*! 
  Inicializace tae pro vcevlknov pouit (kritick sekce). 
  @param[in] uiStart poten hodnota potadla (pednastaven hodnota 0 znamen, e prvn 
  vyprodukovan hodnota pomoc @ref Next bude 1)
  */
  CAutoIncrement(unsigned uiStart = 0);
  //! destruktor
  /*! Uvolnn kritick sekce. */
  ~CAutoIncrement();
  //! hodnota potadla
  /*!
  Postupnm volnm tto metody zskvme vzestupnou hodnotu sel (kadm vykonnm se
  hodnota vnitnho potadla zvtuje). Dokud nedojde k peteen tae, je hodnota
  jedinen.
  @return Souasn hodnota potadla
  */
  unsigned Next(void);
protected:
  //! nastaven
  /*!
  Nastaven hodnoty potadla by se mlo provdt pouze vjimen, nap. pokud zddn tda 
  tak neme provst ji v konstruktoru. Operaci lze provst pouze smrem k vtm hodnotm.
  @param[in] uiValue nov hodnota potadla
  */
  void Set(unsigned uiValue);
private:
  CAutoIncrement(const CAutoIncrement &cSource); // neimplementovno
  CAutoIncrement &operator=(const CAutoIncrement &cRight); // neimplementovno
  unsigned m_uiCurrent;
  CRITICAL_SECTION m_rLock;
};

#endif
