#ifndef _BACKUP_H
#define _BACKUP_H

#include <iosfwd>
#include <string>

/*!
@file Backup.h
Deklarace pomocnch td, kter se pouvaj pi zlohovn databze a pi jejm nslednm obnoven.
*/

//! hlavika seku binrnch dat
/*!
Tda exportuje vtov typ, kter uruje seznam poloek, kter tvo samostatn blok v binrnch
datech zlohy a zrove metodu, kterou se vytv hlavika pslunho bloku.
*/
class CBkHeader {
public:
  //! druhy blok v binrnch datech zlohy
  enum EObject {eDatabase, eCities, eAliases, ePlans, eBuses, eCity, eAlias, ePlan, eBus, eSeat, eEnd};
  //! hodnota hlaviky bloku
  /*!
  Vytvoen hodnoty v hlavice bloku binrnch dat, kter reprezentuje dan druh.
  @param[in] eObject druh objektu uren vtovm typem
  @return hodnota pouit v hlavice pro dan objekt
  */
  unsigned HeaderValue(EObject eObject);
};

//! pomocn tda pro zlohu databze
/*!
Pomoc tto tdy se provd zloha jednotlivch objekt databze do binrnho proudu. Obsahuje
metody pro jednotliv primitivn typy, sloitj se pak poskldaj z tchto stavebnch kamen.
Viz @ref backup.
*/
class CBackup : public CBkHeader {
public:
  //! konstruktor
  /*!
  @param[in] cStream vstupn proud pro proveden zlohy
  */
  CBackup(std::ostream *cStream);
  //! hlavika bloku
  /*!
  Vytvoen hlaviky pro odpovdajc druh objektu a jej zapsn do vstupnho proudu.
  @param[in] eObject druh objektu uren vtovm typem
  */
  void Header(EObject eObject);
  //! zloha logick hodnoty
  /*!
  Hodnota typu @c bool je zapsna do vstupnho proudu.
  @param[in] fValue logick promnn
  */
  void Bool(bool fValue);
  //! zloha neznamnkov hodnoty
  /*!
  Hodnota typu @c unsigned je zapsna do vstupnho proudu.
  @param[in] uiValue neznamnkov celoseln promnn
  */
  void Unsigned(unsigned uiValue);
  //! zloha bloku dat
  /*!
  Do vstupnho proudu je uloen blok binrnch dat (nap. jednoduch struktura).
  @param[in] pAddress adresa bloku dat v pamti
  @param[in] uiSize velikost dat (v bytech)
  */
  void Block(const void *pAddress, unsigned uiSize);
  //! zloha etzce
  /*!
  Do vstupnho proudu je uloen znakov etezec (jazyka C).
  @param[in] pszValue zdrojov etzec
  */
  void String(const char *pszValue);
  //! zloha etzce
  /*!
  Do vstupnho proudu je uloen znakov etezec (tda STL).
  @param[in] strValue zdrojov etzec
  */
  void String(const std::string &strValue);
private:
  void CheckStream(void);
  std::ostream *m_cStream;
};

//! pomocn tda pro obnoven databze ze zlohy
/*!
Pomoc tto tdy se provd obnova jednotlivch objekt databze z binrnho proudu. Obsahuje
metody pro jednotliv primitivn typy, sloitj se pak poskldaj z tchto stavebnch kamen.
Viz @ref backup.
*/
class CRestore : public CBkHeader {
public:
  //! konstruktor
  /*!
  @param[in] cStream vstupn proud se zlohou
  */
  CRestore(std::istream *cStream);
  //! hlavika objektu
  /*!
  Naten a kontrola hlaviky pro poadovan objekt.
  @param[in] eObject druh objektu uren vtovm typem
  */
  void Header(EObject eObject);
  //! naten logick hodnoty
  /*!
  Z proudu je natena logick hodnota.
  @return naten hodnota @c bool
  */
  bool Bool(void);
  //! naten neznamnkov hodnoty
  /*!
  Z proudu je natena celoseln neznamnkov hodnota.
  @return naten hodnota @c unsigned
  */
  unsigned Unsigned(void);
  //! naten bloku dat
  /*!
  Ze vstupnho proudu je naten blok nestrukturovanch dat poadovan velikosti.
  @param[out] pAddress adresa pro uloen bloku dat v pamti
  @param[in] uiSize poadovan velikost dat (v bytech)
  */
  void Block(void *pAddress, unsigned uiSize);
  //! naten etzce
  /*!
  Ze vstupnho proudu jsou nateny znaky etzce.
  @param[out] strValue etzec STL pro uloen znak
  */
  void String(std::string &strValue);
private:
  void CheckStream(void);
  std::istream *m_cStream;
};

#endif
