#include <vector>
#include <list>
#include <set>
#include <algorithm>
#include <time.h>
#include <istream>

#pragma hdrstop

#include "Bus.h"
#include "Seat.h"
#include "Backup.h"
#include "../../Common/Query.h"
#include "../../Common/UtilClass.h"

static const unsigned MAX_READERS = 100;

// funkn objekty - unrn predikty pro remove_if (vrac true, kdy chci vymazat z kontajneru - tj. pi nespchu operace se sedadlem)
static const unsigned ASK_TIMEOUT = 300; // s

// pro doasnou rezervaci sedadel
class CAskSeat : public std::unary_function<unsigned, bool> {
public:
  CAskSeat(std::vector<CSeat> *cSeats, unsigned uiCustomer) : m_cSeats(cSeats), m_uiCustomer(uiCustomer) {}
  bool operator()(unsigned uiSeat) {return uiSeat == 0 || uiSeat > m_cSeats->size() || !m_cSeats->at(--uiSeat).ReserveFor(m_uiCustomer, ASK_TIMEOUT);}
private:
  std::vector<CSeat> *m_cSeats;
  unsigned m_uiCustomer;
};

// pro trvalou rezervaci sedadel
class CLockSeat : public std::unary_function<unsigned, bool> {
public:
  CLockSeat(std::vector<CSeat> *cSeats, unsigned uiCustomer) : m_cSeats(cSeats), m_uiCustomer(uiCustomer) {}
  bool operator()(unsigned uiSeat) {return uiSeat == 0 || uiSeat > m_cSeats->size() || !m_cSeats->at(--uiSeat).Reserve(m_uiCustomer);}
private:
  std::vector<CSeat> *m_cSeats;
  unsigned m_uiCustomer;
};

// pro zruen rezervace sedadel
class CUnlockSeat : public std::unary_function<unsigned, bool> {
public:
  CUnlockSeat(std::vector<CSeat> *cSeats, unsigned uiCustomer) : m_cSeats(cSeats), m_uiCustomer(uiCustomer) {}
  bool operator()(unsigned uiSeat) {return uiSeat == 0 || uiSeat > m_cSeats->size() || !m_cSeats->at(--uiSeat).Free(m_uiCustomer);}
private:
  std::vector<CSeat> *m_cSeats;
  unsigned m_uiCustomer;
};

// ------------- CBus -------------------------------------------------

CBus::CBus(const RBusInfo *prInfo, unsigned uiId, unsigned uiSeats) : RBusInfo(*prInfo), std::vector<CSeat>(uiSeats), CReadWriteLock(MAX_READERS)
{
  CBus::uiId = uiId;
  m_fEnabled = true;
  m_uiSeats = uiSeats;
}

CBus::CBus(std::istream *cStream) : CReadWriteLock(MAX_READERS)
{
  CRestore cRestore(cStream);
  cRestore.Header(CBkHeader::eBus);
  cRestore.Block(static_cast<RBusInfo *>(this), sizeof(RBusInfo));
  m_fEnabled = cRestore.Bool();
  m_uiSeats = cRestore.Unsigned();
  reserve(m_uiSeats); // alokace potebnho prostoru
  for(unsigned i = 0; i < m_uiSeats; ++i)
    push_back(CSeat(cStream));
}

CBus::~CBus()
{
}

void CBus::GetInfo(RBusInfo *prData)
{
  *prData = *static_cast<RBusInfo *>(this);
}

void CBus::Enable(void)
{
  EnterWrite();
  m_fEnabled = true;
  LeaveWrite();
}

void CBus::Disable(void)
{
  EnterWrite();
  m_fEnabled = false;
  LeaveWrite();
}

bool CBus::Enabled(void)
{
  bool fResult;

  EnterRead();
  fResult = m_fEnabled;
  LeaveRead();
  return fResult;
}

void CBus::FreeSeats(std::list<unsigned> *cSeats)
{
  cSeats->clear();
  // naplnn seznamu sly volnch sedadel
  EnterRead();
  for(unsigned i = 0; i < m_uiSeats; ++i) {
    if(!at(i).Reserved())
      cSeats->push_back(i + 1);
  }
  LeaveRead();
}

bool CBus::AskSeats(std::list<unsigned> *cSeats, unsigned uiCustomer)
{
  unsigned uiInSize;
  bool fResult;

  cSeats->sort();
  cSeats->unique(); // odstrann duplicitnch sedadel
  EnterWrite();
  if(m_fEnabled) {
    uiInSize = static_cast<unsigned>(cSeats->size());
    // Prochzme poadovan sedadla, na kadm pomoc funknho objektu provedeme operaci. 
    // Pokud se nepoda, prvek je ze seznamu vymazn.
    cSeats->remove_if(CAskSeat(static_cast<std::vector<CSeat> *>(this), uiCustomer));
    fResult = cSeats->size() == uiInSize; // true, pokud se operace podaila pro vechna sedadla
  }
  else { // autobus nem povoleny rezervace
    cSeats->clear();
    fResult = false;
  }
  LeaveWrite();
  return fResult;
}

bool CBus::LockSeats(std::list<unsigned> *cSeats, unsigned uiCustomer)
{
  unsigned uiInSize;
  bool fResult;

  cSeats->sort();
  cSeats->unique(); // odstrann duplicitnch sedadel
  EnterWrite();
  if(m_fEnabled) {
    uiInSize = static_cast<unsigned>(cSeats->size());
    // Prochzme poadovan sedadla, na kadm pomoc funknho objektu provedeme operaci. 
    // Pokud se nepoda, prvek je ze seznamu vymazn.
    cSeats->remove_if(CLockSeat(static_cast<std::vector<CSeat> *>(this), uiCustomer));
    fResult = cSeats->size() == uiInSize; // true, pokud se operace podaila pro vechna sedadla
  }
  else { // autobus nem povoleny rezervace
    cSeats->clear();
    fResult = false;
  }
  LeaveWrite();
  return fResult;
}

bool CBus::UnlockSeats(std::list<unsigned> *cSeats, unsigned uiCustomer)
{
  unsigned uiInSize;
  bool fResult;

  cSeats->sort();
  cSeats->unique(); // odstrann duplicitnch sedadel
  EnterWrite();
  if(m_fEnabled) {
    uiInSize = static_cast<unsigned>(cSeats->size());
    // Prochzme poadovan sedadla, na kadm pomoc funknho objektu provedeme operaci. 
    // Pokud se nepoda, prvek je ze seznamu vymazn.
    cSeats->remove_if(CUnlockSeat(static_cast<std::vector<CSeat> *>(this), uiCustomer));
    fResult = cSeats->size() == uiInSize; // true, pokud se operace podaila pro vechna sedadla
  }
  else { // autobus nem povoleny rezervace
    cSeats->clear();
    fResult = false;
  }
  LeaveWrite();
  return fResult;
}

void CBus::Backup(std::ostream *cStream) const
{
  CBackup cBackup(cStream);
  cBackup.Header(CBkHeader::eBus);
  cBackup.Block(static_cast<const RBusInfo *>(this), sizeof(RBusInfo));
  cBackup.Bool(m_fEnabled);
  cBackup.Unsigned(m_uiSeats);
  for_each(begin(), end(), [=](const CSeat &cSeat){cSeat.Backup(cStream);});
}

// ------------- CBuses -----------------------------------------------
bool BusIdLess(const CBus &cBus1, const CBus &cBus2)
{
  return cBus1.Id() < cBus2.Id();
}

CBuses::CBuses(void) : std::set<CBus, FnBusIdLess>(BusIdLess)
{
}

CBuses::CBuses(std::istream *cStream) : std::set<CBus, FnBusIdLess>(BusIdLess)
{
  CRestore cRestore(cStream);
  cRestore.Header(CBkHeader::eBuses);
  Set(cRestore.Unsigned()); // potadlo
  unsigned uiItems = cRestore.Unsigned(); // poet poloek
  for(unsigned i = 0; i < uiItems; ++i)
    insert(CBus(cStream));
}

CBuses::~CBuses()
{
}

const CBus *CBuses::Add(const RBusInfo *prInfo, unsigned uiSeats)
{
  return &(*insert(CBus(prInfo, Next(), uiSeats)).first); // vloen autobusu s uniktnm slem
}

bool CBuses::Delete(unsigned uiId)
{
  RBusInfo rInfo;

  return erase(CBus(&rInfo, uiId, 0)) != 0;
}

const CBus *CBuses::Find(unsigned uiId)
{
  RBusInfo rInfo;
  std::set<CBus, FnBusIdLess>::iterator i = find(CBus(&rInfo, uiId, 0));
  return i == end() ? NULL : &(*i); // vrac ukazatel na nalezenou tdu
}

unsigned CBuses::GetNextId(unsigned uiId)
{
  RBusInfo rInfo;
  std::set<CBus, FnBusIdLess>::iterator i = upper_bound(CBus(&rInfo, uiId, 0));
  return i == end() ? 0 : i->Id();
}

unsigned CBuses::Count(void) const
{
  return static_cast<unsigned>(size());
}

void CBuses::Clear(void)
{
  clear();
}

void CBuses::Backup(std::ostream *cStream)
{
  CBackup cBackup(cStream);
  cBackup.Header(CBkHeader::eBuses);
  cBackup.Unsigned(Next()); // potadlo
  cBackup.Unsigned(static_cast<unsigned>(size())); // poet poloek
  for_each(begin(), end(), [=](const CBus &cBus){cBus.Backup(cStream);});
}
