#ifndef _BUS_H
#define _BUS_H

#include <vector>
#include <list>
#include <memory>
#include <set>
#include <iosfwd>
#include <time.h>
#include "Seat.h"
#include "../../Common/Query.h"
#include "../../Common/UtilClass.h"

/*!
@file Bus.h
Deklarace tdy @ref CBus obsahujc kolekci sedadel a tdy @ref CBuses obsahujc kolekci autobus.
*/

//! autobus
/*!
Tda obsahuje vechna data o jednom autobusu. Obsahuje konstatn data ve struktue @ref RBusInfo,
kontajner sedadel (instanc tdy @ref CSeat) a dal poloky (jako zmek pro vcensobn pstup
zddn od @ref CReadWriteLock), kter mohou v prbhu ivota tdy sv hodnoty mnit. Kad instance 
m sv uniktn slo. Informace o poten a clov stanici jsou uloeny jako odkazy na tdy 
@ref CCity, pln sedadel je uren ablonou typu @ref CPlan.

@image html Bus.png "Schematick znzornn tdy CBus"

@note Nkter metody pracuj se seznamy, ve kterch vystupuj sla sedadel. Stejn jako ve skutenm 
autobusu jsou sedadla slovna od 1 (0 nen platn slo sedadla). Nicmn je jasn, e intern jsou
poloky kontajneru sedadel indexovny od 0.
*/
class CBus : private RBusInfo, private std::vector<CSeat>, private CReadWriteLock {
public:
  //! konstruktor
  /*!
  Vytvoen novho autobusu, kter obsahuje kontajner zadanho potu sedadel.
  @param[in] prInfo ukazatel na strukturu @ref RBusInfo, kter obsahuje data pro tento autobus
  @param[in] uiId jedinen ID tohoto autobusu
  @param[in] uiSeats poet sedadel
  */
  CBus(const RBusInfo *prInfo, unsigned uiId, unsigned uiSeats);
  //! naten z proudu
  /*!
  Tda je vytvoena natenm vech poloek z proudu.
  @param[in] cStream vstupn proud s binrnmi daty
  */
  CBus(std::istream *cStream);
  //! destruktor
  ~CBus();
  //! vrt ID autobusu
  unsigned Id(void) const {return uiId;}
  //! vrt ID potenho msta
  unsigned CityFrom(void) const {return uiCityFrom;}
  //! vrt ID clovho msta
  unsigned CityTo(void) const {return uiCityTo;}
  //! vrt ID plnu autobusu
  unsigned Plan(void) const {return uiPlan;}
  //! vrt as odjezdu autobusu
  time_t LeaveTime(void) const {return tLeaveTime;}
  //! vrt poet sedadel v autobusu
  unsigned Seats(void) const {return m_uiSeats;}
  //! voln sedadla
  /*!
  Napln seznam sly volnch sedadel.
  @param[out] cSeats seznam, do kterho se vlo sla volnch sedadel
  @note Ppadn poten obsah pedanho seznamu bude ped naplnnm smazn.
  */
  void FreeSeats(std::list<unsigned> *cSeats);
  //! info o autobusu
  /*!
  Napln strukturu @ref RBusInfo daty o autobusu
  @param[out] prData struktura @ref RBusInfo s daty o autobusu
  */
  void GetInfo(RBusInfo *prData);
  //! povolit rezervace pro autobus
  void Enable(void);
  //! zakzat rezervace pro autobus
  void Disable(void);
  //! dotaz zda je autobus povolen
  /*!
  @return @e true v ppad, e je autobus povolen, @e false v ppad, e je zakzn
  */
  bool Enabled(void);
  //! doasn rezervace
  /*!
  Doasn rezervace sedadel ze seznamu. Sedadla, kter se povedlo rezervovat, zstanou v seznamu, ostatn budou smazna.
  @param[in,out] cSeats seznam sedadel k rezervaci
  @param[in] uiCustomer ID rezervujcho
  @return @e true v ppad vech spnch rezervac, @e false v ppad njakho nezdaru pi rezervaci
  */
  bool AskSeats(std::list<unsigned> *cSeats, unsigned uiCustomer);
  //! trval rezervace
  /*!
  Trval rezervace sedadel ze seznamu. Sedadla, kter se povedlo rezervovat, zstanou v seznamu, ostatn budou smazna.
  @param[in,out] cSeats seznam sedadel k rezervaci
  @param[in] uiCustomer ID rezervujcho
  @return @e true v ppad vech spnch rezervac, @e false v ppad njakho nezdaru pi rezervaci
  */
  bool LockSeats(std::list<unsigned> *cSeats, unsigned uiCustomer);
  //! zruen rezervace
  /*!
  Zruen rezervace sedadel ze seznamu. Sedadla, kter se povedlo uvolnit, zstanou v seznamu, ostatn budou smazna.
  @param[in,out] cSeats seznam sedadel k uvolnn
  @param[in] uiCustomer ID rezervujcho
  @return @e true v ppad vech spnch zruen rezervac, @e false v ppad njakho nezdaru
  */
  bool UnlockSeats(std::list<unsigned> *cSeats, unsigned uiCustomer);
  //! zloha
  /*!
  Vytvoen zlohy tdy vech datovch poloek do proudu (souboru, soketu, ...).
  @param[out] cStream vstupn proud oteven v binrnm mdu
  */
  void Backup(std::ostream *cStream) const;
private:
  bool m_fEnabled;
  unsigned m_uiSeats;
};

//! porovnvac funkce pro kontejner autobus
/*! Porovnv se podle ID autobus */
typedef bool (*FnBusIdLess)(const CBus &, const CBus &);

//! kontajner autobus
/*!
Tda sdruuje vechny autobusy. Star se o sprvn pidvn, mazn a hledn autobus,
dky @ref CAutoIncrement i o jejich jedinen slovn. Autobusy jsou uloeny v asociativnm
kontajneru, azen je provdno dle uniktnho sla.
*/
class CBuses : private std::set<CBus, FnBusIdLess>, private CAutoIncrement {
public:
  //! konstruktor
  /*!
  Vytvoen przdnho kontajneru.
  */
  CBuses(void);
  //! naten z proudu
  /*!
  Kontajner je vytvoen natenm vech poloek z proudu.
  @param[in] cStream vstupn proud s binrnmi daty
  */
  CBuses(std::istream *cStream);
  //! destruktor
  ~CBuses();
  //! pidn novho autobusu
  /*!
  Zskn uniktnho ID pro autobus pomoc @ref CAutoIncrement a vloen do kontejneru.
  @param[in] prInfo data potebn k vytvoen autobusu
  @param[in] uiSeats poet sedadel autobusu
  @return ukazatel na nov vloen autobus
  */
  const CBus *Add(const RBusInfo *prInfo, unsigned uiSeats);
  //! smazn autobusu
  /*!
  @param[in] uiId uniktn ID autobusu urenho ke smazn
  @return @e true pi spchu, @e false jinak
  */
  bool Delete(unsigned uiId);
  //! nalezen autobusu podle ID
  /*!
  @param[in] uiId ID hledanho autobusu
  @return ukazatel na nalezen autobus nebo @e NULL
  */
  const CBus *Find(unsigned uiId);
  //! nalezen dalho autobusu v poad
  /*!
  Nalezne a vrt ID dalho autobusu v poad po @e uiId. Pokud takov neexistuje, vrt @e 0.
  */
  unsigned GetNextId(unsigned uiId);
  //! zjitn potu
  /*!
  Zjitn potu objekt v kontajneru.
  @return aktuln poet
  */
  unsigned Count(void) const;
  //! vymazn kontajneru
  /*!
  Vymazn vech poloek v kontajneru.
  @note Vymazny jsou pouze prvky kontajneru, pokud k nmu existuje njak index, je nutn
  jeho vymazn externmi prostedky.
  */
  void Clear(void);
  //! zloha
  /*!
  Vytvoen zlohy kontajneru vetn dalch datovch poloek do proudu (souboru, soketu, ...).
  @param[out] cStream vstupn proud oteven v binrnm mdu
  */
  void Backup(std::ostream *cStream);
};

#endif
