#ifndef _CITY_H
#define _CITY_H

#include <set>
#include <string>
#include <iosfwd>
#include "../../Common/Query.h"
#include "../../Common/UtilClass.h"

/*!
@file City.h
Deklarace tdy msta @ref CCity a tdy @ref CCities obsahujc kontejner mst a zajiujc prci s nimi.
*/

//! msto
/*!
Tda si krom konstantnch poloek (@ref RCityInfo) uchovv i poet autobus, ve kterch vystupuje jako 
poten a clov stanice.
*/
class CCity : private RCityInfo, private CReadWriteLock {
public:
  //! konstruktor
  /*!
  @param[in] prInfo struktura @ref RCityInfo s daty k vytvoen msta
  */
  CCity(const RCityInfo *prInfo);
  //! naten z proudu
  /*!
  Tda je vytvoena natenm vech poloek z proudu.
  @param[in] cStream vstupn proud s binrnmi daty
  */
  CCity(std::istream *cStream);
  //! destruktor
  ~CCity();
  //! zskn informac o mst
  /*!
  @param[out] prInfo struktura @ref RCityInfo, kter bude naplnna daty o mstu
  */
  void GetInfo(RCityInfo *prInfo);
  //! vrt ID msta
  unsigned Id(void) const {return uiId;}
  //! vrt etzec s nzvem msta
  const char *Name(void) const {return strName.c_str();}
  //! vrt poet odjdjcch autobus
  unsigned LinesOut(void);
  //! zmn poet odjdjcch autobus
  /*!
  @param[in] uiChange hodnota o kterou se m zmnit potadlo odjdjcch autobus
  @note Pro odetn autobusu se pouije zporn hodnota.
  */
  void LinesOut(int uiChange);
  //! vrt poet pijdjcch autobus
  unsigned LinesIn(void);
  //! zmn poet pijdjcch autobus
  /*!
  @param[in] uiChange hodnota o kterou se m zmnit potadlo pijdjcch autobus
  @note Pro odetn autobusu se pouije zporn hodnota.
  */
  void LinesIn(int uiChange);
  //! zloha
  /*!
  Vytvoen zlohy tdy vech datovch poloek do proudu (souboru, soketu, ...).
  @param[out] cStream vstupn proud oteven v binrnm mdu
  */
  void Backup(std::ostream *cStream) const;
private:
  unsigned m_uiLinesOut;
  unsigned m_uiLinesIn;
};

//! porovnvac funkce pro kontejner mst
/*! Porovnv se podle ID msta */
typedef bool (*FnCityIdLess)(const CCity &, const CCity &);

//! kontajner mst
/*!
Tda sdruuje vechna msta a dky @ref CAutoIncrement zajiuje piazovn jedinench sel.
Msta jsou uloena v asociativnm kontajneru, kde jsou azena podle uniktnho sla.
*/
class CCities : private std::set<CCity, FnCityIdLess>, private CAutoIncrement {
public:
  //! konstruktor
  /*!
  Vytvoen przdnho kontajneru.
  */
  CCities(void);
  //! naten z proudu
  /*!
  Kontajner je vytvoen natenm vech poloek z proudu.
  @param[in] cStream vstupn proud s binrnmi daty
  */
  CCities(std::istream *cStream);
  //! destruktor
  ~CCities();
  //! pidn msta
  /*!
  Pid nov msto podle nzvu a piad mu sv slo ID.
  @param[in] pszName nzev msta
  @return ukazatel na nov pidan msto
  */
  const CCity *Add(const char *pszName);
  //! smazn msta
  /*!
  Smae msto podle jeho ID.
  @param[in] uiId ID msta urenho ke smazn
  @return @e true pi spchu, @e false pi selhn
  */
  bool Delete(unsigned uiId);
  //! nalezen msta
  /*!
  Najde msto podle jeho ID a vrt na nj ukazatel.
  @param[in] uiId ID hledanho msta
  @return ukazatel na nalezen msto nebo @e NULL
  */
  const CCity *Find(unsigned uiId);
  //! nalezen dalho msta v poad
  /*!
  Nalezne a vrt ID dalho msta v poad po @e uiId. Pokud takov neexistuje, vrt @e 0.
  */
  unsigned GetNextId(unsigned uiId);
  //! zjitn potu
  /*!
  Zjitn potu objekt v kontajneru.
  @return aktuln poet
  */
  unsigned Count(void) const;
  //! vymazn kontajneru
  /*!
  Vymazn vech poloek v kontajneru.
  @note Vymazny jsou pouze prvky kontajneru, pokud k nmu existuje njak index, je nutn
  jeho vymazn externmi prostedky.
  */
  void Clear(void);
  //! zloha
  /*!
  Vytvoen zlohy kontajneru vetn dalch datovch poloek do proudu (souboru, soketu, ...).
  @param[out] cStream vstupn proud oteven v binrnm mdu
  */
  void Backup(std::ostream *cStream);
};

#endif
