#include <windows.h>
#include <process.h>
#include <string>
#include <istream>
#include <fstream>
#include <stdio.h>

#pragma hdrstop

#include "Journal.h"
#include "MemoryDb.h"
#include "../../Common/Command.h"

static const char pszDbName[] = "brs.database";
static const char pszTempDbName[] = "brs.tmp";
static const char pszJournalName[] = "brs.journal";
static const unsigned uiJournalHeader = 0x6c6e724a;

const char *InternalBackupName(void)
{
  return pszDbName;
}

// --- tda pro zpracovn souboru urnlu

CReadJournal::CReadJournal(const char *pszPath, CMemoryDb *cDb) : m_strPath(pszPath), m_cDb(cDb)
{
}

CReadJournal::~CReadJournal()
{
}

static void MakeSeatList(std::list<unsigned> *cSeats, const char *pcData, unsigned uiLength)
{
  while(uiLength >= sizeof(unsigned)) {
    cSeats->push_back(*(unsigned *)pcData);
    pcData += sizeof(unsigned);
    uiLength -= sizeof(unsigned);
  }
}

bool CReadJournal::Replay(unsigned *puiItems)
{
  unsigned uiHeader;
  unsigned uiCommand;
  unsigned uiLength;
  char *pcData;

  *puiItems = 0;
  std::ifstream cJournal;
  cJournal.open(m_strPath + pszJournalName, std::ifstream::binary);
  if(cJournal.fail())
    return true; // dn urnl
  bool fResult = true;
  while(fResult && cJournal.peek() != std::ifstream::traits_type::eof()) {
    // naten poloky ze urnlu
    cJournal.read((char *)&uiHeader, sizeof(unsigned));
    cJournal.read((char *)&uiCommand, sizeof(unsigned));
    cJournal.read((char *)&uiLength, sizeof(unsigned));
    // naten dat poloky
    if(cJournal.good() && uiLength != 0) {
      pcData = new char [uiLength];
      cJournal.read(pcData, uiLength);
    }
    else
      pcData = NULL;
    // zpracovn poloky
    if(!cJournal.fail() && uiHeader == uiJournalHeader) {
      switch(uiCommand) {
        case CCommand::CMDADDCITY:
          fResult = m_cDb->AddCity(pcData) != 0;
          break;
        case CCommand::CMDADDCITYALIAS:
          {
            unsigned uiCityOffset = static_cast<unsigned>(strlen(pcData) + 1);
            fResult = m_cDb->AddCityAlias(pcData, *(unsigned *)(pcData + uiCityOffset)) != 0;
          }
          break;
        case CCommand::CMDDELETECITY:
          fResult = m_cDb->DeleteCity(pcData);
          break;
        case CCommand::CMDADDBUSPLAN:
          {
            unsigned uiSeatsOffset = static_cast<unsigned>(strlen(pcData) + 1);
            unsigned uiStrPlanOffset = uiSeatsOffset + sizeof(unsigned);
            unsigned uiBinPlanOffset = uiStrPlanOffset + static_cast<unsigned>(strlen(pcData + uiStrPlanOffset) + 1);
            CBlobEx<std::string> cBlob((const char *)(pcData + uiStrPlanOffset), (const void *)(pcData + uiBinPlanOffset), uiLength - uiBinPlanOffset);
            fResult = m_cDb->AddBusPlan(pcData, *(unsigned *)(pcData + uiSeatsOffset), std::move(&cBlob)) != 0;
          }
          break;
        case CCommand::CMDDELETEBUSPLAN:
          fResult = m_cDb->DeleteBusPlan(pcData);
          break;
        case CCommand::CMDADDBUS:
          fResult = uiLength == sizeof(RBusInfo) && m_cDb->AddBus((const RBusInfo *)pcData) != 0;
          break;
        case CCommand::CMDDELETEBUS:
          fResult = uiLength == sizeof(unsigned) && m_cDb->DeleteBus(*(unsigned *)pcData);
          break;
        case CCommand::CMDENABLEBUS:
          fResult = uiLength == sizeof(unsigned) && m_cDb->EnableBus(*(unsigned *)pcData);
          break;
        case CCommand::CMDDISABLEBUS:
          fResult = uiLength == sizeof(unsigned) && m_cDb->DisableBus(*(unsigned *)pcData);
          break;
        case CCommand::CMDNEWCUSTOMER:
          if(uiLength == 0)
            m_cDb->NewCustomer();
          else
            fResult = false;
          break;
        case CCommand::CMDLOCKSEATS:
          {
            unsigned uiSeatsOffset = sizeof(unsigned) + sizeof(unsigned);
            if(uiLength >= uiSeatsOffset) {
              std::list<unsigned> cSeats;
              MakeSeatList(&cSeats, pcData + uiSeatsOffset, uiLength - uiSeatsOffset);
              fResult = m_cDb->LockSeats(*(unsigned *)pcData, &cSeats, *(unsigned *)(pcData + sizeof(unsigned)));
            }
            else
              fResult = false;
          }
          break;
        case CCommand::CMDUNLOCKSEATS:
          {
            unsigned uiSeatsOffset = sizeof(unsigned) + sizeof(unsigned);
            if(uiLength >= uiSeatsOffset) {
              std::list<unsigned> cSeats;
              MakeSeatList(&cSeats, pcData + uiSeatsOffset, uiLength - uiSeatsOffset);
              fResult = m_cDb->UnlockSeats(*(unsigned *)pcData, &cSeats, *(unsigned *)(pcData + sizeof(unsigned)));
            }
            else
              fResult = false;
          }
          break;
        default: fResult = false; break; // nepovolen nebo neznm pkaz
      }
      (*puiItems)++;
    }
    else
      fResult = false;
    delete [] pcData;
  }
  cJournal.close();
  if(!fResult)
    rename((m_strPath + pszJournalName).c_str(), (m_strPath + pszJournalName + "_failed").c_str()); // zloha nezpracovanho urnlu
  return fResult;
}

// --- tda pro zpis do souboru urnlu a automatick zlohy

CWriteJournal::CWriteJournal(const char *pszPath, unsigned uiLimit, CMemoryDb *cDb)
  : m_cDb(cDb), m_strPath(pszPath), m_uiLimit(uiLimit)
{
  InitializeCriticalSection(&m_rLock);
  // oteven souboru urnlu
  m_cFile.open(m_strPath + pszJournalName, std::ofstream::binary | std::ofstream::app);
  // synchronizan udlosti (s klesajc prioritou)
  m_ahEvent[0] = CreateEvent(NULL, TRUE, FALSE, NULL); // backup - nesignalizovan manual-reset udlost
  m_ahEvent[1] = CreateEvent(NULL, TRUE, FALSE, NULL); // terminate - nesignalizovan manual-reset udlost
  // vytvoen pracovnho vlkna
  m_hThread = (HANDLE)_beginthreadex(NULL, 0, WorkThread, this, 0, NULL);
}

CWriteJournal::~CWriteJournal()
{
  SetEvent(m_ahEvent[1]); // poadavek na ukonen
  WaitForSingleObject(m_hThread, INFINITE); // ekej na dokonen vlkna
  // destrukce alokovanch zdroj
  DeleteCriticalSection(&m_rLock);
  CloseHandle(m_hThread);
  CloseHandle(m_ahEvent[0]);
  CloseHandle(m_ahEvent[1]);
  // uzaven urnlu
  m_cFile.close();
}

void CWriteJournal::AskBackup(void)
{
  SetEvent(m_ahEvent[0]); // nastaven poadavku na proveden zlohy
}

bool CWriteJournal::AddItem(unsigned uiType, unsigned uiLength, const void *pData)
{
  EnterCriticalSection(&m_rLock); // zajitn serializace poadavk
  m_cFile.write((const char *)&uiJournalHeader, sizeof(unsigned)); // hlavika
  m_cFile.write((const char *)&uiType, sizeof(unsigned)); // typ
  m_cFile.write((const char *)&uiLength, sizeof(unsigned)); // dlka dat
  if(uiLength != 0)
    m_cFile.write((const char *)pData, uiLength); // data
  m_cFile.flush(); // vyprzdnn vstupnho bufferu - zapsn na disk
  if(static_cast<unsigned>(m_cFile.tellp()) > m_uiLimit)
    SetEvent(m_ahEvent[0]); // nastaven poadavku na proveden zlohy
  LeaveCriticalSection(&m_rLock);
  return m_cFile.good();
}

void CWriteJournal::InternalBackup(void)
{
  if(m_cDb != NULL && m_cFile.tellp() > 0) {
    m_cDb->EnterWrite(); // zmek cel databze - vlun pstup
    EnterCriticalSection(&m_rLock); // zajitn serializace poadavk
    m_cFile.close();
    // proveden zlohy databze
    rename((m_strPath + pszDbName).c_str(), (m_strPath + pszTempDbName).c_str()); // doasn zloha
    std::ofstream cBackup;
    cBackup.open(m_strPath + pszDbName, std::ofstream::binary | std::ofstream::trunc);
    bool fDone = m_cDb->BackupNoLock(&cBackup); // zajiuje vhradn pstup k databzi
    cBackup.close();
    if(fDone) {
      remove((m_strPath + pszTempDbName).c_str()); // vymazn doasn zlohy
      // vymazn (znovuoteven) urnlu
      m_cFile.open(m_strPath + pszJournalName, std::ofstream::binary | std::ofstream::trunc);
    }
    LeaveCriticalSection(&m_rLock);
    m_cDb->LeaveWrite(); // zmek cel databze
  }
  ResetEvent(m_ahEvent[0]); // poadavek je vyzen
}

unsigned int WINAPI CWriteJournal::WorkThread(LPVOID pThreadParams)
{
  CWriteJournal *cThis = (CWriteJournal *)pThreadParams;
  while(true) {
    switch(WaitForMultipleObjects(2, cThis->m_ahEvent, FALSE, INFINITE)) {
      case WAIT_OBJECT_0 + 0: // proveden zlohy databze
        cThis->InternalBackup();
        break;
      case WAIT_OBJECT_0 + 1: // ukonen
        return 0;
    }
  }
}
