#ifndef _JOURNAL_H
#define _JOURNAL_H

#include <windows.h>
#include <iosfwd>
#include <fstream>
#include <string>

class CMemoryDb;

/*!
@file Journal.h
Zde jsou deklarovny tdy pro podporu urnlu databze.
*/

//! jmno souboru intern zlohy databze
/*!
Funkce, kter zpstupuje jmno souboru pro intern zlohu databze.
@return jmno souboru jako etzec jazyka C
*/
extern const char *InternalBackupName(void);

//! tda pro obnoven dat z urnlu
/*!
S pomoc stedn metody @ref Replay je mon provst obnoven dat, kter se (patrn 
vlivem selhn hardware) nedostala do intern databze.
*/
class CReadJournal {
public:
  //! konstruktor
  /*!
  Vytvoen instance tdy s nastavenm parametr.
  @param[in] pszPath sloka, ve kter se bude hledat soubor urnlu
  @param[in] cDb ukazatel na instanci databze
  */
  CReadJournal(const char *pszPath, CMemoryDb *cDb);
  //! destruktor
  ~CReadJournal();
  //! penesen poloek urnlu do databze
  /*!
  Hlavn metoda tdy slou k penesen vech poloek ze urnlu (tj. tch, kter se dve 
  nedostaly do kompletn intern zlohy databze). Postupn jsou jednotliv poloky vyzvedvny 
  ze souboru, jsou vyzvednuty a sprvn konvertovny jednotliv parametry a nsledn se volaj 
  pslun metody databze pro proveden poadovanch innost. Pi kadm voln databzov 
  metody se kontroluje nvratov hodnota.
  @param[out] puiItems uruje poet zpracovanch poloek. Ten v ppad spn zpracovanho 
  urnlu je roven celkovmu potu poloek, v ppad nespchu odkazuje na poloku, pi jejm 
  zpracovn dolo k chyb.
  @return spch operace. Hodnota @e true pro spn zpracovn celho souboru urnlu nebo 
  v ppad, e soubor urnlu zcela chyb nebo je przdn. Hodnota @e false zna chybu - 
  patn formt souboru nebo nespch konkrtn databzov operace.
  */
  bool Replay(unsigned *puiItems);
private:
  std::string m_strPath;
  CMemoryDb *m_cDb;
};

//! tda pro zpis do urnlu
/*!
Tda zajiuje prci se souborem urnlu. Pro svou prci si vytvo pomocn vlkno (je je 
po naprostou vtinu asu uspno). Slou k pidvn jednotlivch poloek do urnlu, 
zrove kontroluje, zda velikost souboru nepeshla pednastaven limit. Pokud ano, je 
probuzeno pracovn vlkno, kter zajist proveden zlohy databze a vyprzdnn urnlu.
Podrobnji @ref journal "zde".
*/
class CWriteJournal {
public:
  //! konstruktor
  /*!
  Vytvoen tdy obn i vytvoen pracovnho vlkna pro provdn zlohy databze
  (vetn synchronizanm objekt) a vytvoen (przdnho) souboru urnlu v poadovanm 
  umstn.
  @param[in] pszPath uren sloky pro uloen intern zlohy databze a souboru urnlu.
  Konkrtn jmna jsou urena vnitnmi konstantami. Volajc je odpovdn za to, e 
  proces m odpovdajc pstupov prva do zadan sloky. 
  @param[in] uiLimit limit velikosti urnlovho souboru. Pi jeho pekroen je vyvolna 
  intern zloha databze.
  @param[in] cDb ukazatel na instanci databze, se kterou je urnl spojen.
  */
  CWriteJournal(const char *pszPath, unsigned uiLimit, CMemoryDb *cDb);
  //! destruktor
  /*!
  Ped zruenm tdy je legln ukoneno pracovn vlkno a jsou uvolnny synchronizan 
  objekty.
  */
  ~CWriteJournal();
  //! poadavek na vyprzdnn urnlu
  /*!
  Obsluha vnjho poadavku na proveden intern zlohy cel databze a tm i vyprzdnn 
  souboru urnlu. Poadavek je uloen a bude vyzen pracovnm vlknem co nejdve (po 
  zskn vhradnho pstupu k databzi).
  */
  void AskBackup(void);
  //! pidn poloky do urnlu
  /*!
  Do urnlu je pidna dal poloka dle parametr metody. V ppad, e velikost urnlu 
  pekro pednastavenou mez, bude co nejdve provedena kompletn zloha cel databze a 
  soubor urnlu nsledn vymazn.
  @param[in] uiType typ poloky (pro ely identifikace poloky se pouv vtov typ
  @ref CCommand::EQueryCmd)
  @param[in] uiLength velikost dat (v bytech) 
  @param[in] pData ukazatel na data, kter jsou spojena s konkrtnm pkazem (velikost a 
  struktura dat se li podle pkazu)
  @return spch operace
  */
  bool AddItem(unsigned uiType, unsigned uiLength, const void *pData);
private:
  void InternalBackup(void);
  std::string m_strPath;
  unsigned m_uiLimit;
  CMemoryDb *m_cDb;
  std::ofstream m_cFile;
  static unsigned int WINAPI WorkThread(LPVOID pThreadParams);
  HANDLE m_hThread;
  HANDLE m_ahEvent[2];
  CRITICAL_SECTION m_rLock;
};

#endif
