#include <set>
#include <list>
#include <map>
#include <algorithm>
#include <string>
#include <string.h>
#include <istream>

#pragma hdrstop

#include "../../Common/Query.h"
#include "../Admin.h"
#include "Backup.h"
#include "Bus.h"
#include "MemoryDb.h"
#include "Journal.h"
#include "../../Common/UtilClass.h"
#include "../../Common/Command.h"

// porovnavn mst dle jmna - azen v indexu
bool CityLess(const CCity *cCity1, const CCity *cCity2)
{
  return strcmp(cCity1->Name(), cCity2->Name()) < 0;
}

// porovnavn pln dle jmna - azen v indexu
bool PlanLess(const CPlan *cPlan1, const CPlan *cPlan2)
{
  return strcmp(cPlan1->Name(), cPlan2->Name()) < 0;
}

// porovnvn autobus dle poten a clov stanice a data odjezdu - azen v indexu
bool BusLess(const CBus *cBus1, const CBus *cBus2)
{
  if(cBus1->CityFrom() < cBus2->CityFrom())
    return true;
  else if(cBus1->CityFrom() == cBus2->CityFrom() && cBus1->CityTo() < cBus2->CityTo())
    return true;
  else if(cBus1->CityFrom() == cBus2->CityFrom() && cBus1->CityTo() == cBus2->CityTo() && cBus1->LeaveTime() < cBus2->LeaveTime())
    return true;
  else
    return false;
}

/*
Tdy CCity, CPlan, CBus se pouvaj v asociativnch kontajnerech (pmo nebo pomoc ukazatel). 
Metody jako CCities::Add, CCities::Find a analogicky pro ostatn vrac konstantn ukazatele na poadovan objekt.
V implementacch metod ne se nkolikrt pouv const_cast takto zskanho ukazatele, protoe se nadle
pouv k voln nekonstatnch metod objekt CCity atd., kter jsou v tchto kontajnerech uloeny. Nein to pote, 
protoe ve tdch CCity atd. lze mnit pouze poloky, kter nemaj vliv na uspodn a tud nemohou naruit 
konzistenci dat v kontajneru ani v pslunch indexech. Konstatn poloky jsou soustedny do struktur RCityInfo, 
RPlanInfo a RBusInfo a tyto se mimo konstruktor ji nikde nemn.
*/

static const unsigned MAX_READERS = 100;
static const unsigned JOURNAL_LIMIT = 10 * 1024;

CMemoryDb::CMemoryDb(void) : CReadWriteLock(MAX_READERS)
{
  Init();
  // przdn databze
  m_cCities = new CCities; // msta
  m_cCityAlias = new std::map<std::string, unsigned>; // nhradn jmna pro msta
  m_cPlans = new CPlans; // plny autobus
  m_cBuses = new CBuses; // autobusy
}

CMemoryDb::CMemoryDb(std::istream *cStream) : CReadWriteLock(MAX_READERS)
{
  Init();
  // kontajnery budou nateny z proudu
  m_cCities = NULL; // msta
  m_cCityAlias = NULL; // nhradn jmna pro msta
  m_cPlans = NULL; // plny autobus
  m_cBuses = NULL; // autobusy
  // naten z proudu v chrnnm bloku - vyvoln vjimky pi selhn kterkoliv podoperace
  try {
    CRestore cRestore(cStream);
    // hlavika
    cRestore.Header(CBkHeader::eDatabase);
    Set(cRestore.Unsigned()); // potadlo zkaznk
    // msta
    m_cCities = new CCities(cStream);
    // aliasy mst
    m_cCityAlias = new std::map<std::string, unsigned>;
    cRestore.Header(CBkHeader::eAliases);
    unsigned uiCount = cRestore.Unsigned();
    for(unsigned i = 0; i < uiCount; ++i) {
      cRestore.Header(CBkHeader::eAlias);
      std::string strAlias;
      cRestore.String(strAlias);
      unsigned uiId = cRestore.Unsigned();
      m_cCityAlias->insert(std::make_pair(strAlias, uiId));
    }
    // plny
    m_cPlans = new CPlans(cStream);
    // autobusy
    m_cBuses = new CBuses(cStream);
    // konec
    cRestore.Header(CBkHeader::eEnd);
  }
  catch(std::ios::failure &) {
    // vytvoen neexistujcch nebo vyprzdnn ji vytvoench kontajner v ppad chyby
    if(m_cCities == NULL) // msta
      m_cCities = new CCities;
    else
      m_cCities->Clear();
    if(m_cCityAlias == NULL) // nhradn jmna pro msta
      m_cCityAlias = new std::map<std::string, unsigned>;
    else
      m_cCityAlias->clear();
    if(m_cPlans == NULL) // plny autobus
      m_cPlans = new CPlans;
    else
      m_cPlans->Clear();
    if(m_cBuses == NULL) // autobusy
      m_cBuses = new CBuses;
    else
      m_cBuses->Clear();
  }
  unsigned uiId = 0;
  // vytvoen index mst
  while((uiId = NextCity(uiId)) != 0)
    m_cCityIndex->insert(const_cast<CCity *>(m_cCities->Find(uiId)));
  // vytvoen index pln
  while((uiId = NextPlan(uiId)) != 0)
    m_cPlanIndex->insert(const_cast<CPlan *>(m_cPlans->Find(uiId)));
  // vytvoen index autobus
  while((uiId = NextBus(uiId)) != 0)
    m_cBusIndex->insert(const_cast<CBus *>(m_cBuses->Find(uiId)));
}

CMemoryDb::~CMemoryDb()
{
  delete m_cLockCities;
  delete m_cLockPlans;
  delete m_cLockBuses;
  delete m_cBuses;
  delete m_cBusIndex;
  delete m_cPlans;
  delete m_cPlanIndex;
  delete m_cCities;
  delete m_cCityIndex;
  delete m_cCityAlias;
  if(m_cJournal != NULL) {
    m_cJournal->AskBackup();
    Sleep(0);
    delete m_cJournal;
  }
}

void CMemoryDb::Init(void)
{
  m_cLockCities = new CReadWriteLock(MAX_READERS);
  m_cLockPlans = new CReadWriteLock(MAX_READERS);
  m_cLockBuses = new CReadWriteLock(MAX_READERS);
  m_cCityIndex = new std::set<CCity *, FnCityLess>(CityLess); // index mst
  m_cPlanIndex = new std::set<CPlan *, FnPlanLess>(PlanLess); // index pln
  m_cBusIndex = new std::set<CBus *, FnBusLess>(BusLess); // index autobus
  m_cJournal = NULL;
}

unsigned CMemoryDb::AddCity(const char *pszName)
{
  unsigned uiCity;
  CCity *cCity;

  EnterRead(); // zmek cel databze
  m_cLockCities->EnterWrite(); // subsystm mst
  uiCity = FindCityNoLock(pszName); // nalezen msta dle jmna
  if(uiCity == 0) {
    // pro nov msto vytvoen instance tdy v kontajneru
    cCity = const_cast<CCity *>(m_cCities->Add(pszName));
    // pidn odkazu do indexu
    m_cCityIndex->insert(cCity);
    uiCity = cCity->Id();
    if(m_cJournal != NULL) // zpis zmny databze do urnlu
      m_cJournal->AddItem(CCommand::CMDADDCITY, static_cast<unsigned>(strlen(pszName) + 1), pszName);
  }
  m_cLockCities->LeaveWrite(); // subsystm mst
  LeaveRead(); // zmek cel databze
  return uiCity;
}

bool CMemoryDb::AddCityAlias(const char *pszName, unsigned uiId)
{
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockCities->EnterWrite(); // subsystm mst
  if(FindCityNoLock(pszName) != 0 || m_cCities->Find(uiId) == NULL)
    fResult = false; // msto pod jmnem i aliasem ji existuje nebo neexistuje podle sla
  else { // vloen novho aliasu
    fResult = m_cCityAlias->insert(std::make_pair(pszName, uiId)).second;
    if(fResult && m_cJournal != NULL) { // zpis zmny databze do urnlu
      unsigned uiStrLen = static_cast<unsigned>(strlen(pszName) + 1);
      unsigned uiSize = uiStrLen + sizeof(unsigned);
      char *pcData = new char [uiSize];
      memcpy(pcData, pszName, uiStrLen);
      memcpy(pcData + uiStrLen, &uiId, sizeof(unsigned));
      m_cJournal->AddItem(CCommand::CMDADDCITYALIAS, uiSize, pcData);
      delete [] pcData;
    }
  }
  m_cLockCities->LeaveWrite(); // subsystm mst
  LeaveRead(); // zmek cel databze
  return fResult;
}

bool CMemoryDb::DeleteCity(const char *pszName)
{
  unsigned uiCity;
  RCityInfo rInfo;
  unsigned uiLinesOut;
  unsigned uiLinesIn;
  bool fResult;

  fResult = false;
  EnterRead(); // zmek cel databze
  m_cLockCities->EnterWrite(); // subsystm mst
  uiCity = FindCityNoLock(pszName); // nalezen msta dle jmna
  if(uiCity != 0) {
    // kontrola, zda msto nen vyuito nkterm autobusem
    CityInfoNoLock(uiCity, &rInfo, &uiLinesOut, &uiLinesIn);
    if(uiLinesOut == 0 && uiLinesIn == 0) {
      m_cCityIndex->erase(&CCity(&rInfo)); // vymazn z indexu
      m_cCities->Delete(uiCity); // vymazn instance tdy
      // vymazn alias (linern sloitost)
      for(std::map<std::string, unsigned>::iterator i = m_cCityAlias->begin(); i != m_cCityAlias->end(); ) {
        if(i->second == uiCity)
          m_cCityAlias->erase(i++);
        else
          ++i;
      }
      fResult = true;
      if(m_cJournal != NULL) // zpis zmny databze do urnlu
        m_cJournal->AddItem(CCommand::CMDDELETECITY, static_cast<unsigned>(strlen(pszName) + 1), pszName);
    }
  }
  m_cLockCities->LeaveWrite(); // subsystm mst
  LeaveRead(); // zmek cel databze
  return fResult;
}

unsigned CMemoryDb::FindCityNoLock(const char *pszName)
{
  RCityInfo rInfo;
  std::set<CCity *, FnCityLess>::iterator i;
  std::map<std::string, unsigned>::iterator j;

  rInfo.uiId = 0; // nen vyuito
  rInfo.strName = pszName;
  i = m_cCityIndex->find(&CCity(&rInfo)); // hledej v indexu dle jmna
  if(i != m_cCityIndex->end())
    return (*i)->Id();
  else {
    j = m_cCityAlias->find(pszName); // jinak prohledme jet aliasy
    return j == m_cCityAlias->end() ? 0 : j->second;
  }
}

unsigned CMemoryDb::FindCity(const char *pszName)
{
  unsigned uiResult;

  EnterRead(); // zmek cel databze
  m_cLockCities->EnterRead(); // subsystm mst
  uiResult = FindCityNoLock(pszName);
  m_cLockCities->LeaveRead(); // subsystm mst
  LeaveRead(); // zmek cel databze
  return uiResult;
}

bool CMemoryDb::CityInfoNoLock(unsigned uiCity, RCityInfo *prInfo, unsigned *puiLinesOut, unsigned *puiLinesIn)
{
  CCity *cCity;

  cCity = const_cast<CCity *>(m_cCities->Find(uiCity)); // nalezen dle sla
  if(cCity == NULL)
    return false;
  else {
    if(prInfo != NULL)
      cCity->GetInfo(prInfo); // konstatn informace
    if(puiLinesOut != NULL)
      *puiLinesOut = cCity->LinesOut(); // jako poten stanice linky
    if(puiLinesIn != NULL)
      *puiLinesIn = cCity->LinesIn(); // jako koncov stanice linky
    return true;
  }
}

bool CMemoryDb::CityInfo(unsigned uiCity, RCityInfo *prInfo, unsigned *puiLinesOut, unsigned *puiLinesIn)
{
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockCities->EnterRead(); // subsystm mst
  fResult = CityInfoNoLock(uiCity, prInfo, puiLinesOut, puiLinesIn);
  m_cLockCities->LeaveRead(); // subsystm mst
  LeaveRead(); // zmek cel databze
  return fResult;
}

unsigned CMemoryDb::AddBusPlan(const char *pszName, unsigned uiSeats, const CBlobEx<std::string> *cData)
{
  unsigned uiPlan;
  CPlan *cPlan;

  EnterRead(); // zmek cel databze
  m_cLockPlans->EnterWrite(); // subsystm pln
  uiPlan = FindBusPlanNoLock(pszName); // nalezen plnu dle jmna
  if(uiPlan == 0) {
    // pro nov pln vytvoen instance tdy v kontajneru
    cPlan = const_cast<CPlan *>(m_cPlans->Add(pszName, uiSeats, cData));
    // pidn odkazu do indexu
    m_cPlanIndex->insert(cPlan);
    uiPlan = cPlan->Id();
    if(m_cJournal != NULL) { // zpis zmny databze do urnlu
      unsigned uiStrLen = static_cast<unsigned>(strlen(pszName) + 1);
      unsigned uiStrPlanLen = static_cast<unsigned>(cData->Get().length() + 1);
      unsigned uiSize = uiStrLen + sizeof(unsigned) + uiStrPlanLen + cData->DataLen();
      char *pcData = new char [uiSize];
      unsigned uiOffset = 0;
      memcpy(pcData + uiOffset, pszName, uiStrLen);
      uiOffset += uiStrLen;
      memcpy(pcData + uiOffset, &uiSeats, sizeof(unsigned));
      uiOffset += sizeof(unsigned);
      memcpy(pcData + uiOffset, cData->Get().c_str(), uiStrPlanLen);
      uiOffset += uiStrPlanLen;
      memcpy(pcData + uiOffset, cData->DataPointer(), cData->DataLen());
      m_cJournal->AddItem(CCommand::CMDADDBUSPLAN, uiSize, pcData);
      delete [] pcData;
    }
  }
  m_cLockPlans->LeaveWrite(); // subsystm pln
  LeaveRead(); // zmek cel databze
  return uiPlan;
}

bool CMemoryDb::DeleteBusPlan(const char *pszName)
{
  unsigned uiPlan;
  RPlanInfo rInfo;
  unsigned uiBuses;
  bool fResult;

  fResult = false;
  EnterRead(); // zmek cel databze
  m_cLockPlans->EnterWrite(); // subsystm pln
  uiPlan = FindBusPlanNoLock(pszName); // nalezen plnu dle jmna
  if(uiPlan != 0) {
    // kontrola, zda pln nen vyuit nkterm autobusem
    BusPlanInfoNoLock(uiPlan, &rInfo, &uiBuses, NULL);
    if(uiBuses == 0) {
      m_cPlanIndex->erase(&CPlan(&rInfo, NULL)); // vymazn z indexu
      m_cPlans->Delete(uiPlan); // vymazn instance tdy
      fResult = true;
      if(m_cJournal != NULL) // zpis zmny databze do urnlu
        m_cJournal->AddItem(CCommand::CMDDELETEBUSPLAN, static_cast<unsigned>(strlen(pszName) + 1), pszName);
    }
  }
  m_cLockPlans->LeaveWrite(); // subsystm pln
  LeaveRead(); // zmek cel databze
  return fResult;
}

unsigned CMemoryDb::FindBusPlanNoLock(const char *pszName)
{
  RPlanInfo rInfo;
  std::set<CPlan *, FnPlanLess>::iterator i;

  rInfo.uiId = 0; // nen vyuito
  rInfo.uiSeats = 0; // nen vyuito
  rInfo.strName = pszName;
  i = m_cPlanIndex->find(&CPlan(&rInfo, NULL)); // hledej v indexu dle jmna
  return i == m_cPlanIndex->end() ? 0 : (*i)->Id();
}

unsigned CMemoryDb::FindBusPlan(const char *pszName)
{
  unsigned uiResult;

  EnterRead(); // zmek cel databze
  m_cLockPlans->EnterRead(); // subsystm pln
  uiResult = FindBusPlanNoLock(pszName);
  m_cLockPlans->LeaveRead(); // subsystm pln
  LeaveRead(); // zmek cel databze
  return uiResult;
}

bool CMemoryDb::BusPlanInfoNoLock(unsigned uiPlan, RPlanInfo *prInfo, unsigned *puiBuses, CBlobEx<std::string> *cData)
{
  CPlan *cPlan;

  cPlan = const_cast<CPlan *>(m_cPlans->Find(uiPlan)); // nalezen dle sla
  if(cPlan == NULL)
    return false;
  else {
    if(prInfo != NULL)
      cPlan->GetInfo(prInfo); // konstatn informace
    if(puiBuses != NULL)
      *puiBuses = cPlan->Buses(); // poet autobus s tmto plnem
    if(cData != NULL)
      cPlan->Data(cData); // binrn data
    return true;
  }
}

bool CMemoryDb::BusPlanInfo(unsigned uiPlan, RPlanInfo *prInfo, unsigned *puiBuses, CBlobEx<std::string> *cData)
{
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockPlans->EnterRead(); // subsystm pln
  fResult = BusPlanInfoNoLock(uiPlan, prInfo, puiBuses, cData);
  m_cLockPlans->LeaveRead(); // subsystm pln
  LeaveRead(); // zmek cel databze
  return fResult;
}

unsigned CMemoryDb::AddBus(const RBusInfo *prBus)
{
  CBus *cBus;
  unsigned uiBus;
  CCity *cCityFrom;
  CCity *cCityTo;
  CPlan *cPlan;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterWrite(); // subsystm autobus
  uiBus = FindBusNoLock(prBus->uiCityFrom, prBus->uiCityTo, prBus->tLeaveTime); // hledn v indexu
  if(uiBus == 0 || m_cBuses->Find(uiBus)->LeaveTime() != prBus->tLeaveTime) { // takov autobus dosud neexistuje
    // kontrola, zda vchoz i clov msto a pln existuj
    m_cLockCities->EnterRead(); // subsystm mst
    m_cLockPlans->EnterRead(); // subsystm pln
    cCityFrom = const_cast<CCity *>(m_cCities->Find(prBus->uiCityFrom));
    cCityTo = const_cast<CCity *>(m_cCities->Find(prBus->uiCityTo));
    cPlan = const_cast<CPlan *>(m_cPlans->Find(prBus->uiPlan));
    if(cCityFrom != NULL && cCityTo != NULL && cPlan != NULL) {
      cBus = const_cast<CBus *>(m_cBuses->Add(prBus, cPlan->Seats())); // vytvoen novho autobusu
      m_cBusIndex->insert(cBus); // aktualizace indexu
      // aktualizace dalch dat
      cCityFrom->LinesOut(+1);
      cCityTo->LinesIn(+1);
      cPlan->Buses(+1);
      uiBus = cBus->Id();
      if(m_cJournal != NULL) // zpis zmny databze do urnlu
        m_cJournal->AddItem(CCommand::CMDADDBUS, sizeof(RBusInfo), prBus);
    }
    m_cLockPlans->LeaveRead(); // subsystm pln
    m_cLockCities->LeaveRead(); // subsystm mst
  }
  m_cLockBuses->LeaveWrite(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return uiBus;  
}

bool CMemoryDb::DeleteBus(unsigned uiBus)
{
  CBus *cBus;
  CCity *cCity;
  CPlan *cPlan;
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterWrite(); // subsystm autobus
  cBus = const_cast<CBus *>(m_cBuses->Find(uiBus)); // nalezen autobusu dle sla
  if(cBus == NULL)
    fResult = false;
  else {
    // aktualizace dat v dalch kontajnerech (poloky by vdy mly existovat)
    m_cLockCities->EnterRead(); // subsystm mst
    cCity = const_cast<CCity *>(m_cCities->Find(cBus->CityFrom()));
    if(cCity != NULL)
      cCity->LinesOut(-1);
    cCity = const_cast<CCity *>(m_cCities->Find(cBus->CityTo()));
    if(cCity != NULL)
      cCity->LinesIn(-1);
    m_cLockCities->LeaveRead(); // subsystm mst
    m_cLockPlans->EnterRead(); // subsystm pln
    cPlan = const_cast<CPlan *>(m_cPlans->Find(cBus->Plan()));
    if(cPlan != NULL)
      cPlan->Buses(-1);
    m_cLockPlans->LeaveRead(); // subsystm pln
    // zruen instance tdy i aktualizace indexu
    m_cBusIndex->erase(m_cBusIndex->find(cBus));
    m_cBuses->Delete(uiBus);
    fResult = true;
    if(m_cJournal != NULL) // zpis zmny databze do urnlu
      m_cJournal->AddItem(CCommand::CMDDELETEBUS, sizeof(unsigned), &uiBus);
  }
  m_cLockBuses->LeaveWrite(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return fResult;
}

unsigned CMemoryDb::FindBusNoLock(unsigned uiFrom, unsigned uiTo, time_t tLeaveTime)
{
  RBusInfo rInfo;
  std::set<CBus *, FnBusLess>::iterator i;

  // hledn v indexu dle stanic a asu odjezdu
  rInfo.uiId = 0; // nen vyuito
  rInfo.uiCityFrom = uiFrom;
  rInfo.uiCityTo = uiTo;
  rInfo.tLeaveTime = tLeaveTime;
  rInfo.uiLine = 0; // nen vyuito
  rInfo.uiPlan = 0; // nen vyuito
  i = m_cBusIndex->lower_bound(&CBus(&rInfo, 0, 0)); // >= poadovan hodnot
  if(i == m_cBusIndex->end() || (*i)->CityFrom() != uiFrom || (*i)->CityTo() != uiTo)
    return 0; // poten a clov stanice mus souhlasit
  else
    return (*i)->Id(); // as odjezdu me bt stejn, ppadn i vt
}

unsigned CMemoryDb::FindBus(unsigned uiFrom, unsigned uiTo, time_t tLeaveTime)
{
  unsigned uiResult;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterRead(); // subsystm autobus
  uiResult = FindBusNoLock(uiFrom, uiTo, tLeaveTime);
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return uiResult;
}

bool CMemoryDb::BusInfo(unsigned uiBus, RBusInfo *prInfo, bool *pfEnabled, unsigned *puiSeats, std::list<unsigned> *cFreeSeats)
{
  CBus *cBus;
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterRead(); // subsystm autobus
  cBus = const_cast<CBus *>(m_cBuses->Find(uiBus)); // nalezen podle sla
  if(cBus != NULL) {
    if(prInfo != NULL)
      cBus->GetInfo(prInfo); // konstatn informace
    if(pfEnabled != NULL)
      *pfEnabled = cBus->Enabled(); // monost provdn rezervac
    if(puiSeats != NULL)
      *puiSeats = cBus->Seats(); // poet sedadel
    if(cFreeSeats != NULL)
      cBus->FreeSeats(cFreeSeats); // aktuln voln sedadla
    fResult = true;
  }
  else
    fResult = false;
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return fResult;
}

bool CMemoryDb::EnableBus(unsigned uiBus)
{
  CBus *cBus;
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterRead(); // subsystm autobus
  cBus = const_cast<CBus *>(m_cBuses->Find(uiBus));
  if(cBus != NULL) {
    cBus->Enable();
    fResult = true;
    if(m_cJournal != NULL) // zpis zmny databze do urnlu
      m_cJournal->AddItem(CCommand::CMDENABLEBUS, sizeof(unsigned), &uiBus);
  }
  else
    fResult = false;
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return fResult;
}

bool CMemoryDb::DisableBus(unsigned uiBus)
{
  CBus *cBus;
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterRead(); // subsystm autobus
  cBus = const_cast<CBus *>(m_cBuses->Find(uiBus));
  if(cBus != NULL) {
    cBus->Disable();
    fResult = true;
    if(m_cJournal != NULL) // zpis zmny databze do urnlu
      m_cJournal->AddItem(CCommand::CMDDISABLEBUS, sizeof(unsigned), &uiBus);
  }
  else
    fResult = false;
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return fResult;
}

unsigned CMemoryDb::NewCustomer(void)
{
  unsigned uiResult;

  EnterRead(); // zmek cel databze
  uiResult = Next(); // nov uniktn slo zkaznka
  if(m_cJournal != NULL) // zpis zmny databze do urnlu
    m_cJournal->AddItem(CCommand::CMDNEWCUSTOMER, 0, NULL);
  LeaveRead(); // zmek cel databze
  return uiResult;
}

bool CMemoryDb::AskSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer)
{
  CBus *cBus;
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterRead(); // subsystm autobus
  cBus = const_cast<CBus *>(m_cBuses->Find(uiBus)); // nalezen autobusu dle sla
  if(cBus == NULL) { // autobus nenalezen
    cSeats->clear(); // vymazn vech sedadel
    fResult = false;
  }
  else
    fResult = cBus->AskSeats(cSeats, uiCustomer); // proveden operace
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return fResult;
}

bool CMemoryDb::LockSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer)
{
  CBus *cBus;
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterRead(); // subsystm autobus
  cBus = const_cast<CBus *>(m_cBuses->Find(uiBus)); // nalezen autobusu dle sla
  if(cBus == NULL) { // autobus nenalezen
    cSeats->clear(); // vymazn vech sedadel
    fResult = false;
  }
  else {
    fResult = cBus->LockSeats(cSeats, uiCustomer); // proveden operace
    if(!cSeats->empty() && m_cJournal != NULL) { // zpis zmny databze do urnlu
      unsigned uiSize = sizeof(unsigned) + sizeof(unsigned) + static_cast<unsigned>(cSeats->size()) * sizeof(unsigned);
      char *pcData = new char [uiSize]; // autobus, zkaznk, sedadla
      unsigned uiOffset = 0;
      memcpy(pcData + uiOffset, &uiBus, sizeof(unsigned));
      uiOffset += sizeof(unsigned);
      memcpy(pcData + uiOffset, &uiCustomer, sizeof(unsigned));
      uiOffset += sizeof(unsigned);
      std::for_each(cSeats->begin(), cSeats->end(), [=, &uiOffset](unsigned &uiSeat){memcpy(pcData + uiOffset, &uiSeat, sizeof(unsigned)); uiOffset += sizeof(unsigned);});
      m_cJournal->AddItem(CCommand::CMDLOCKSEATS, uiSize, pcData);
      delete [] pcData;
    }
  }
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return fResult;
}

bool CMemoryDb::UnlockSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer)
{
  CBus *cBus;
  bool fResult;

  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterRead(); // subsystm autobus
  cBus = const_cast<CBus *>(m_cBuses->Find(uiBus)); // nalezen autobusu dle sla
  if(cBus == NULL) { // autobus nenalezen
    cSeats->clear(); // vymazn vech sedadel
    fResult = false;
  }
  else {
    fResult = cBus->UnlockSeats(cSeats, uiCustomer); // proveden operace
    if(!cSeats->empty() && m_cJournal != NULL) { // zpis zmny databze do urnlu
      unsigned uiSize = sizeof(unsigned) + sizeof(unsigned) + static_cast<unsigned>(cSeats->size()) * sizeof(unsigned);
      char *pcData = new char [uiSize]; // autobus, zkaznk, sedadla
      unsigned uiOffset = 0;
      memcpy(pcData + uiOffset, &uiBus, sizeof(unsigned));
      uiOffset += sizeof(unsigned);
      memcpy(pcData + uiOffset, &uiCustomer, sizeof(unsigned));
      uiOffset += sizeof(unsigned);
      std::for_each(cSeats->begin(), cSeats->end(), [=, &uiOffset](unsigned &uiSeat){memcpy(pcData + uiOffset, &uiSeat, sizeof(unsigned)); uiOffset += sizeof(unsigned);});
      m_cJournal->AddItem(CCommand::CMDUNLOCKSEATS, 0, NULL);
      delete [] pcData;
    }
  }
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return fResult;
}

unsigned CMemoryDb::NextBus(unsigned uiBus)
{
  EnterRead(); // zmek cel databze
  m_cLockBuses->EnterRead(); // subsystm autobus
  unsigned uiResult = m_cBuses->GetNextId(uiBus);
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return uiResult;
}

unsigned CMemoryDb::NextCity(unsigned uiCity)
{
  EnterRead(); // zmek cel databze
  m_cLockCities->EnterRead(); // subsystm mst
  unsigned uiResult = m_cCities->GetNextId(uiCity);
  m_cLockCities->LeaveRead(); // subsystm mst
  LeaveRead(); // zmek cel databze
  return uiResult;
}

unsigned CMemoryDb::NextPlan(unsigned uiPlan)
{
  EnterRead(); // zmek cel databze
  m_cLockPlans->EnterRead(); // subsystm pln
  unsigned uiResult = m_cPlans->GetNextId(uiPlan);
  m_cLockPlans->LeaveRead(); // subsystm pln
  LeaveRead(); // zmek cel databze
  return uiResult;
}

bool CMemoryDb::ListCity(const char *pszFrom, std::string *strName)
{
  EnterRead(); // zmek cel databze
  m_cLockCities->EnterRead(); // subsystm mst
  bool fResult = false;
  RCityInfo rInfo = {0, pszFrom};
  CCity temp(&rInfo);
  auto it = m_cCityIndex->lower_bound(&temp);
  if(it != m_cCityIndex->end()) {
    *strName = (*it)->Name();
    fResult = true;
  }
  m_cLockCities->LeaveRead(); // subsystm mst
  LeaveRead(); // zmek cel databze
  return fResult;
}

bool CMemoryDb::ListAlias(const char *pszFrom, std::string *strAlias, std::string *strCity)
{
  EnterRead(); // zmek cel databze
  m_cLockCities->EnterRead(); // subsystm mst
  bool fResult = false;
  auto it = m_cCityAlias->lower_bound(pszFrom);
  if(it != m_cCityAlias->end()) {
    *strAlias = it->first;
    const CCity* cCity = m_cCities->Find(it->second);
    *strCity = cCity->Name();
    fResult = true;
  }
  m_cLockCities->LeaveRead(); // subsystm mst
  LeaveRead(); // zmek cel databze
  return fResult;
}

// funkn objekt pro uloen aliasu msta
class CBackupAlias {
public:
  CBackupAlias(CBackup *cBackup) : m_cBackup(cBackup) {}
  void operator()(const std::pair<std::string, unsigned> &rData) {m_cBackup->Header(CBkHeader::eAlias); m_cBackup->String(rData.first); m_cBackup->Unsigned(rData.second);}
private:
  CBackup *m_cBackup;
};

bool CMemoryDb::BackupNoLock(std::ostream *cStream)
{
  bool fResult = true;
  // vstup v chrnnm bloku - vyvoln vjimky pi selhn kterkoliv podoperace
  try {
    CBackup cBackup(cStream);
    // hlavika
    cBackup.Header(CBkHeader::eDatabase);
    cBackup.Unsigned(Next()); // potadlo (jedna vynechan hodnota nevad, podstatn je jedinenost)
    // msta
    m_cCities->Backup(cStream);
    // aliasy mst
    cBackup.Header(CBkHeader::eAliases);
    cBackup.Unsigned(static_cast<unsigned>(m_cCityAlias->size())); // poet poloek
    CBackupAlias cBackupAlias(&cBackup);
    for_each(m_cCityAlias->begin(), m_cCityAlias->end(), cBackupAlias);
    // plny
    m_cPlans->Backup(cStream);
    // autobusy
    m_cBuses->Backup(cStream);
    // synchronizace proudu - vyprzdnn vstupnho buferu
    cBackup.Header(CBkHeader::eEnd);
    cStream->flush();
  }
  catch(std::ios::failure &) {
    fResult = false;
  }
  return fResult;
}

bool CMemoryDb::Backup(std::ostream *cStream)
{
  EnterWrite(); // zmek cel databze - vlun pstup
  bool fResult = BackupNoLock(cStream);
  LeaveWrite(); // zmek cel databze
  return fResult;
}

bool CMemoryDb::Info(RDbInfo *prInfo)
{
  if(prInfo == NULL)
    return false;
  EnterRead(); // zmek cel databze
  m_cLockCities->EnterRead(); // subsystm mst
  prInfo->uiCities = m_cCities->Count();
  prInfo->uiCityAliases = static_cast<unsigned>(m_cCityAlias->size());
  m_cLockCities->LeaveRead(); // subsystm mst
  m_cLockPlans->EnterRead(); // subsystm pln
  prInfo->uiPlans = m_cPlans->Count();
  m_cLockPlans->LeaveRead(); // subsystm pln
  m_cLockBuses->EnterRead(); // subsystm autobus
  prInfo->uiBuses = m_cBuses->Count();
  m_cLockBuses->LeaveRead(); // subsystm autobus
  LeaveRead(); // zmek cel databze
  return true;
}

void CMemoryDb::OnStart(void)
{
  if(m_cJournal == NULL)
    m_cJournal = new CWriteJournal(m_strDataDir.c_str(), JOURNAL_LIMIT, this);
}

void CMemoryDb::OnStop(void)
{
  if(m_cJournal != NULL) {
    m_cJournal->AskBackup();
    Sleep(0);
    delete m_cJournal;
    m_cJournal = NULL;
  }
}

void CMemoryDb::SetDataDir(const char *pszDataDir)
{
  m_strDataDir = pszDataDir;
}
