#ifndef _MEMORYDB_H
#define _MEMORYDB_H

#include <set>
#include <list>
#include <map>
#include <string>
#include <iosfwd>
#include "../../Common/Query.h"
#include "../Admin.h"
#include "City.h"
#include "Plan.h"
#include "Bus.h"
#include "../../Common/UtilClass.h"

/*!
@file MemoryDB.h
Zde je deklarovna tda @ref CMemoryDb, kter implementuje virtuln metody tdy @ref CQuery.
*/

//! typ porovnvac funkce pro index mst
typedef bool (*FnCityLess)(const CCity *, const CCity *);
//! typ porovnvac funkce pro index pln autobus
typedef bool (*FnPlanLess)(const CPlan *, const CPlan *);
//! typ porovnvac funkce pro index autobus
typedef bool (*FnBusLess)(const CBus *, const CBus *);

class CWriteJournal;

//! Hlavn tda databze
/*!
Tda obsahuje metody pro prci s autobusy, msty a plny autobus. Implementuje rozhran @ref CQuery
pro klientsk voln a @ref CAdmin pro sprvu
a zapouzduje tak veker operace s databz. Tda vlastn obsahuje jak vlastn databzov stroj, tak
zajiuje uloen dat pomoc kontajner C++ (tedy pmo v operan pamti potae). Krom vlastnch 
datovch kontajner pro msta, plny a autobusy implementuje i pomocn kontajnery (indexy) pro rychl 
vyhledvn podle rznch kritri. Pro snaz vyhledvn mst podporuje i jejich pomocn jmna 
(aliasy).
@image html Db.png "Databze z pohledu jejho uivatele a sprvce"

Zkladn entity, se ktermi databze pracuje, jsou reprezentovny tdami:
@li @ref CCity - msto jako poten a koncov stanice autobusov linky,
@li @ref CPlan - typ autobusu, tj. poet sedadel pro cestujc a jejich rozmstn,
@li @ref CSeat - sedadlo v autobuse a stav jeho rezervace,
@li @ref CBus - autobus, kter sdruuje kolekci sedadel.

Poet sedadel v konkrtnm autobusu je dn jeho plnem a v prbhu ivotnosti je pevn dan, pstup
k jednotlivm sedadlm je pomoc indexu, pro uloen je zvolen <tt>std::vector<CSeat></tt>. Ostatn tdy
jsou uloeny ve stromov struktue, kter zaruuje rychl vyhledvn. Pouitm kontajnerem je
<tt>std::set</tt> (nap. tedy <tt>std::set<CCity></tt>), setdn je podle uniktnho sla. Pro 
vyhledvn podle dodatench kritri (nap. msta dle nzvu nebo autobusu dle koncovch mst a asu 
odjezdu) jsou soust databze pslun indexy, tak kontajnery typu <tt>std::set</tt>, kter vak 
pro sporu msta obsahuj pouze odkazy na hledan objekty a jsou setdny podle poadovanho kritria. 
Aliasy mst jsou implementovny jako asociativn pole pomoc <tt>std::map</tt>.

Objekty @ref CCity, @ref CPlan a @ref CBus obecn nejsou po dobu svho ivota konstantn. Pesto mohou bt
bez problm uloeny v asociativnch kontajnerech, nebo poloky pouit pro uspodn konstantn 
zstvaj (tyto jsou inicializovny v konstruktoru a nsledn ji nen mon je mnit).

Pedpokld se, e k databzi me souasn pistupovat vce klient najednou (tj. voln metod me
probhat souasn z vce vlken). Proto je implementace tdy vybavena zmky typu <i>vce ten/jeden
zpis</i>, kter na jedn stran zajiuj konzistenci dat, na druh se sna minimalizovat prostoje
(podrobnji viz @ref db_lock). 

eenm otzky bezpenosti uloen dat v operan pamti se podrobnji zabv kapitola @ref journal.
*/
class CMemoryDb : public CQuery, public CAdmin, private CAutoIncrement, private CReadWriteLock {
public:
  //! defaultn konstruktor
  /*!
  Vytvoen przdn databze, tj. inicializace vech soust a vytvoen przdnch kontajner
  mst, jejich alias, pln a autobus vetn przdnch index.
  */
  CMemoryDb(void);
  //! konstruktor - obnoven ze zlohy
  /*!
  Databze je vytvoena ve stavu, kter odpovd dve proveden zloze (pomoc metody @ref CMemoryDb::Backup).
  Vechny indexy jsou sprvn synchronizovny s vlastnmi datovmi kontajnery.

  Pokud pi obnov ze zlohy dojde k njak chyb (nesprvn formt dat, chyba pi ten z proudu apod.)
  je vytvoena przdn databze.

  Vce viz @ref backup.
  */
  CMemoryDb(std::istream *cStream);
  //! destruktor
  virtual ~CMemoryDb();
  // rozhran CQuery
  virtual unsigned AddCity(const char *pszName);
  virtual bool AddCityAlias(const char *pszName, unsigned uiId);
  virtual bool DeleteCity(const char *pszName);
  virtual unsigned FindCity(const char *pszName);
  virtual bool CityInfo(unsigned uiCity, RCityInfo *prInfo, unsigned *puiLinesOut, unsigned *puiLinesIn);
  virtual unsigned AddBusPlan(const char *pszName, unsigned uiSeats, const CBlobEx<std::string> *cData);
  virtual bool DeleteBusPlan(const char *pszName);
  virtual unsigned FindBusPlan(const char *pszName);
  virtual bool BusPlanInfo(unsigned uiPlan, RPlanInfo *prInfo, unsigned *puiBuses, CBlobEx<std::string> *cData);
  virtual unsigned AddBus(const RBusInfo *prBus);
  virtual bool DeleteBus(unsigned uiBus);
  virtual unsigned FindBus(unsigned uiFrom, unsigned uiTo, time_t tLeaveTime);
  virtual bool BusInfo(unsigned uiBus, RBusInfo *prInfo, bool *pfEnabled, unsigned *puiSeats, std::list<unsigned> *cFreeSeats);
  virtual bool EnableBus(unsigned uiBus);
  virtual bool DisableBus(unsigned uiBus);
  virtual unsigned NewCustomer(void);
  virtual bool AskSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer);
  virtual bool LockSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer);
  virtual bool UnlockSeats(unsigned uiBus, std::list<unsigned> *cSeats, unsigned uiCustomer);
  virtual unsigned NextBus(unsigned uiBus);
  virtual unsigned NextCity(unsigned uiCity);
  virtual unsigned NextPlan(unsigned uiPlan);
  virtual bool ListCity(const char *pszFrom, std::string *strName);
  virtual bool ListAlias(const char *pszFrom, std::string *strAlias, std::string *strCity);
  // rozhran CAdmin
  virtual bool Backup(std::ostream *cStream);
  virtual bool Info(RDbInfo *prInfo);
  //! akce pi sputn serveru
  /*!
  Provd inicializaci urnlovacho mechanismu, kter zajiuje zachovn dat i v ppad 
  neoekvanho pdu aplikace.
  */
  virtual void OnStart(void);
  //! akce pi zastaven serveru
  /*!
  Po zastaven komunikanch server se provede uloen pamov databze do perzistentn podoby 
  (do diskovho souboru). Tm je zajitno uchovn dat i po ukonen celho serveru.
  */
  virtual void OnStop(void);
  //! adres pro datov soubory
  /*!
  Nastaven adrese pro soubory s intern zlohou databze a urnlu.
  @param[in] pszDataDir cesta k adresi - przdn etzec pro aktuln adres, jinak absolutn 
  nebo relativn cesta zakonen @e backslash.
  */
  void SetDataDir(const char *pszDataDir);
private:
  CCities *m_cCities;
  std::set<CCity *, FnCityLess> *m_cCityIndex;
  std::map<std::string, unsigned> *m_cCityAlias;
  CPlans *m_cPlans;
  std::set<CPlan *, FnPlanLess> *m_cPlanIndex;
  CBuses *m_cBuses;
  std::set<CBus *, FnBusLess> *m_cBusIndex;
  CReadWriteLock *m_cLockCities;
  CReadWriteLock *m_cLockPlans;
  CReadWriteLock *m_cLockBuses;
  void Init(void);
  unsigned FindCityNoLock(const char *pszName);
  bool CityInfoNoLock(unsigned uiCity, RCityInfo *prInfo, unsigned *puiLinesOut, unsigned *puiLinesIn);
  unsigned FindBusPlanNoLock(const char *pszName);
  bool BusPlanInfoNoLock(unsigned uiPlan, RPlanInfo *prInfo, unsigned *puiBuses, CBlobEx<std::string> *cData);
  unsigned FindBusNoLock(unsigned uiFrom, unsigned uiTo, time_t tLeaveTime);
  std::string m_strDataDir;
  CWriteJournal *m_cJournal;
  bool BackupNoLock(std::ostream *cStream);
// sptelen tdy
  //! tda pro zpis do urnlu pi modifikaci dat v databzi
  friend class CWriteJournal;
};

#endif
