#ifndef _PLAN_H
#define _PLAN_H

#include <set>
#include <string>
#include <iosfwd>
#include "../../Common/Query.h"
#include "../../Common/UtilClass.h"

/*!
@file Plan.h
Deklarace tdy @ref CPlan obsahujc pln sedadel autobusu a tdy @ref CPlans obsahujc kolekci pln.
*/

//! pln autobusu
/*!
Tda si krom konstantnch poloek (@ref RPlanInfo) uchovv i poet autobus, kter dan pln pouvaj.
Pln autobusu m textovou a binrn podobu (pro rzn zpsoby pouit).
*/
class CPlan : private RPlanInfo, private CBlobEx<std::string>, private CReadWriteLock {
public:
  //! konstruktor
  /*!
  @param[in] prInfo struktura @ref RPlanInfo se zkladnmi daji
  @param[in] cBlob tda @ref CBlobEx s textovmi a binrnmi daty rozloen sedadel
  */
  CPlan(const RPlanInfo *prInfo, const CBlobEx<std::string> *cBlob);
  //! naten z proudu
  /*!
  Tda je vytvoena natenm vech poloek z proudu.
  @param[in] cStream vstupn proud s binrnmi daty
  */
  CPlan(std::istream *cStream);
  //! destruktor
  ~CPlan();
  //! zskn informac o plnu
  /*!
  Vyplnn struktury @ref RPlanInfo.
  @param[out] prInfo struktura s daty
  */
  void GetInfo(RPlanInfo *prInfo);
  //! vrt ID plnu
  unsigned Id(void) const {return uiId;}
  //! vrt nzev plnu jako konstantn etzec
  const char *Name(void) const {return strName.c_str();}
  //! vrt poet sedadel
  unsigned Seats(void) const {return uiSeats;}
  //! vrt poet autobus pouvajcch tento pln
  unsigned Buses(void);
  //! zmna tae autobus pouvajcch tento pln o hodnotu @a uiChange
  void Buses(int uiChange);
  //! zskn textovch i binrnch informac plnu
  /*!
  @param[out] cData tda obsahujc kopii textovch i binrnch dat plnu
  */
  void Data(CBlobEx<std::string> *cData);
  //! zloha
  /*!
  Vytvoen zlohy tdy vech datovch poloek do proudu (souboru, soketu, ...).
  @param[out] cStream vstupn proud oteven v binrnm mdu
  */
  void Backup(std::ostream *cStream) const;
private:
  unsigned m_uiBuses;
};

//! porovnvac funkce pro kontejner pln
/*! Porovnv se podle ID pln */
typedef bool (*FnPlanIdLess)(const CPlan &, const CPlan &);

//! kontajner pln
/*!
Tda sdruuje vechny plny autobus. Star se o sprvn pidvn, mazn a hledn pln,
dky @ref CAutoIncrement i o jejich jedinen slovn. Plny jsou uloeny v asociativnm 
kontajneru, azen je provedeno dle uniktnho sla.
*/
class CPlans : private std::set<CPlan, FnPlanIdLess>, private CAutoIncrement {
public:
  //! konstruktor
  /*!
  Vytvoen przdnho kontajneru.
  */
  CPlans(void);
  //! naten z proudu
  /*!
  Kontajner je vytvoen natenm vech poloek z proudu.
  @param[in] cStream vstupn proud s binrnmi daty
  */
  CPlans(std::istream *cStream);
  //! destruktor
  ~CPlans();
  //! pidn novho plnu
  /*!
  Zskn uniktnho ID pro pln pomoc @ref CAutoIncrement a vloen do kontejneru.
  @param[in] pszName nzev plnu nebo typu autobusu
  @param[in] uiSeats poet sedadel autobusu
  @param[in] cBlob binrn data rozvren sedadel
  @return ukazatel na nov vloen pln
  */
  const CPlan *Add(const char *pszName, unsigned uiSeats, const CBlobEx<std::string> *cBlob);
  //! smazn plnu
  /*!
  @param[in] uiId uniktn ID plnu urenho ke smazn
  @return @e true pi spchu, @e false jinak
  */
  bool Delete(unsigned uiId);
  //! nalezen plnu podle ID
  /*!
  @param[in] uiId ID hledanho plnu
  @return ukazatel na nalezen pln nebo @e NULL
  */
  const CPlan *Find(unsigned uiId);
  //! nalezen dalho plnu v poad
  /*!
  Nalezne a vrt ID dalho plnu v poad po @e uiId. Pokud takov neexistuje, vrt @e 0.
  */
  unsigned GetNextId(unsigned uiId);
  //! zjitn potu
  /*!
  Zjitn potu objekt v kontajneru.
  @return aktuln poet
  */
  unsigned Count(void) const;
  //! vymazn kontajneru
  /*!
  Vymazn vech poloek v kontajneru.
  @note Vymazny jsou pouze prvky kontajneru, pokud k nmu existuje njak index, je nutn
  jeho vymazn externmi prostedky.
  */
  void Clear(void);
  //! zloha
  /*!
  Vytvoen zlohy kontajneru vetn dalch datovch poloek do proudu (souboru, soketu, ...).
  @param[out] cStream vstupn proud oteven v binrnm mdu
  */
  void Backup(std::ostream *cStream);
};

#endif
